--
--    Copyright (c) 2016 by SAS Institute Inc., Cary, NC, USA.
-- 

--- The SAS Risk and Finance Workbench API provides a Lua-based interface that you can use to programmatically 
-- perform tasks in SAS Risk and Finance Workbench. The Lua programming language is provided with SAS, and you
-- can use the LUA procedure to run Lua code from within a SAS program.
-- @module rfw_api
local M={} 
require 'risk.common.global_init'
local utils    = require 'risk.common.rest_utils'
local Json     = require 'risk.common.json'
local metadata = require 'risk.common.metadata'
local rest     = require 'risk.stress.rest_service'
local alloc    = require 'risk.stress.rfw_allocation'
local str_util = require 'risk.common.string_utils'

Json.strictTypes = true
local rfw_url
local root_dir
M.testing = false
 
--- Check argument types. This is an internal function.
-- @param args a variable list of arguments. The first argument is a table that contains the types.
-- @return If the evaluated arguments are of the correct type, a value of TRUE is returned. If not, 
-- an error message is generated that describes the type mismatch.
-- @usage check_args(ccvs, key, workgroups,{[1]="string",[2]="number", [3]={"nil","table"}, fn="foo"})
local function check_args(...)
    local arg = {...}
     -- find largest index as nils break ipairs
     local max = 0
     for i,v in pairs(arg) do
         if i > max then max = i end
     end
    if max < 2 then return false end
    local types = arg[max]
    max = max -1
    local a
    for i = 1,max do
       a = arg[i]
       if type(types[i]) == "string" then
          assert(type(a) == types[i], "Expecting argument "..i.." to function " .. types.fn .. " to be a " .. types[i] .. " but found " .. type(a) .. ".")
       else
          local matches = false
          for j, t in ipairs(types[i]) do
             if type(a) == t then matches = true end
          end
          assert(matches, "Expecting argument "..i.." to function " .. types.fn .. " to be one of " .. table.tostring(types[i]) .. " but found " .. type(a) .. ".") 
       end       
    end
    return true
end

--[[****************************************************
    *                         #TOC                     *
    ****************************************************

    #UTILITY
    #WORKSHEET
    #DATA
    #PROJECT
    #PROCESS
    #DIMENSION
    #HIERARCHY
    #SCENARIO
    #MODEL
    #MODEL GROUP
    #MODEL GROUP RUN
    #SCRIPT
    #ENTITY
    #REPORT
    #WORKGROUP
    #TASK
    #DETAIL DATA
    
]]--
 
 

--[[****************************************************
   *                    #UTILITY                       *
   *************************************************]]--
   
--- Utility Functions
-- @section Utility


--- Set the REST URL for SAS Risk and Finance Workbench, or reset the URL, if it was already provided. If no URL is provided, 
--  the function attempts to derive the URL from metadata.
--  @param url[opt] the URL for the SAS Risk and Finance Workbench REST API. For example: http://www.example.com/SASRiskandFinanceWorkbench/rest/. 
--  @return a Boolean value that indicates whether the URL can be accessed. A value of true indicates success.
function M.url_set(url)
   Debug("--- url_set --- ")
   if url == nil and rfw_url == nil then
       -- try to get it from metadata
       -- First get the rest api information from metadata
        local hostname, port, protocol, appName
        hostname, port, protocol, appName, rfw_url = metadata.get_rest_end_point_dtls("SASRiskandFinanceWorkbench REST")
        if not rfw_url then
--            sas.print("%1zUnable to get the metadata information. Check that the server is running and that you have the proper permissions.")
           sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_set_1', 'NOQUOTE'))
           return false
         end
        
   elseif url ~= nil then
      rfw_url = url .. '/SASRiskandFinanceWorkbench/rest/';
   end
   if rest.url ~= rfw_url then
      Debug("rest service URL in use is",rest.url,"which does not match rfw_url - setting it to: ",rfw_url)
      local pass = rest.set_url(rfw_url)
      if not pass then
   --        sas.print("%1zInvalid host name or port <%s>.", rfw_url )
          sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_set_2', 'NOQUOTE'), rfw_url )
          return false
      end
   end
   return true
end


--- Set the root directory that SAS Risk and Finance Workbench uses for detail data.
--  @param root_path[opt] the full path to the detail data directory. If the value is `nil`, the path is derived from metadata
--  using the `root_dir_get` function.
--  @return a Boolean value that indicates whether the directory was set. A value of true indicates success.
function M.rootdir_set(root_path)
   Debug("--- rootdir_set --- ")
    root_dir = root_path or M.root_dir_get( )
    return true
end

--- Retrieve the metadata information that is used to derive the root directory for SAS Risk and Finance Workbench.
--  @return `root_dir`: the full path to the root directory.
function M.root_dir_get( )
   Debug("--- root_dir_get --- ")
  local meta_stuff = metadata.get_properties_from_soft_comp("Risk Finance Wkbnch Svr Cfg")
  local root_dir
  for i, ps in ipairs( meta_stuff) do
    if ps.property_name == "root.dir" then
      root_dir = ps.defaultvalue
    end
  end

  if root_dir == nil then
--     sas.print("%1zCan't find root directory ." )
    sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_root_dir_get_1', 'NOQUOTE' ))
    return
  end

  Debug("Root directory=", root_dir)
  
  return root_dir
end

--- Retrieve the directory path for the staging area. This information is derived from the
--  SAS Risk Workgroups API.
--  @param workgroup[OPTIONAL] - string, Name of the workgroup to fetch the path for. Default is public.
--  @return `dir`: the full path to the staging area directory. Nil if call failed.
function M.staging_dir_get(workgroup)
   Trace("staging_dir_get",workgroup)
   check_args( workgroup,{["fn"]="staging_dir_get",{'string','nil'} })
   
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return 
   end
   
   workgroup = (workgroup or 'public'):lower()
   local groups = M.user_workgroups_get()
   if not groups then
      return
   end
   for _,v in ipairs(groups) do
      if v.name:lower() == workgroup then
         return tostring(v.location)..'/staging'
      end      
   end
   return
end

--- Retrieve the ID stored in the database that is associated with a file in the staging area. 
--  @param file_name - string, the name of the file, including the file extension. If the file is not directly in the staging directory, but rather
--               in a sub-directory, the relative path to the file is required.
--               Example1: file is in public workgroup, in staging area. Simply pass 'file_name.ext'
--               Example2: file is in public workgroup, in a subfolder within the staging directory called sub1. Pass 'sub1/file_name.ext'
--               Example3: file is in admin workgroup, in a subfolder called sub1. Pass 'sub1/file_name.ext' and pass the value 'admin' as the
--                         workgroup arg.
--  @param workgroup[OPTIONAL] - string, the name of the workgroup for the file. Default is 'public'
--  @return ID - number, the ID for the specified file. Nil if the id was not found.
--  @return pass - boolean, true if call was successful (without error). Note that the call can be successful but no matching file name is found.
function M.file_id_get(file_name, workgroup)
   Trace("file_id_get",file_name, workgroup)
   check_args( file_name, workgroup,{["fn"]="file_id_get", "string",{'string','nil'} })
   
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return nil, false 
   end
   
   -- Workgroup defaults to public. Make lowercase to match against typelessSurrogatePath, where it will also be made
   -- lowercase, since workgroups are not case sensitive
   workgroup = (workgroup or 'public'):lower()
   -- Construct path to match against typelessSurrogatePath, which only uses '/'. This also gets rid of the the double slash that results if the user puts a / or \
   -- in front of the file name.
   local full_file_name = (workgroup..'/'..file_name):gsub('([\\/]+)','/')

   --Get all files in the server staging areas
   local header = "Accept: application/json, text/javascript, */*; q=0.01 \r\nAccept-Encoding: gzip, deflate, sdch"
   local pass, json_out, code = rest.request( "get", "files", nil, header, nil )
   if not pass then
--       sas.print("%1z Unable to get the file IDs from the staging area: "..code.." "..msg)
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_file_id_get_1', 'NOQUOTE'),tostring(code),json_out)
      return nil, false
   end
   local files = Json:decode(json_out)
   for i,v in pairs(files.items) do
      --typelessSurrogatePath has the form: workgroup/relative_file_path. workgroup is not case sensitive, however.
      --So, before comparing, have to strip out the workgroup and make it lowercase, then attach the rest of the path
      -- This next bit finds the first '/', takes everything to the left, makes that all lowercase, then attaches the rest back on to the right.
      local first_slash_char = v.typelessSurrogatePath:find('/')
      local compare_path = v.typelessSurrogatePath:sub(1,first_slash_char-1):lower()..v.typelessSurrogatePath:sub(first_slash_char)
      if compare_path == full_file_name then
         return v.id, true
      end
   end
   return nil, false
end

--- Retrieve the metadata details for an object with a specified key.
--  @param md the metadata object for a specified key, as obtained from a GET request.
--  @param key the name of the key to retrieve metadata for.
--  @return A value that contains the metadata details. If the key is not found, `nil` is returned.
function M.metadata_find_entry(md, key)
    local tab = md.properties
    for i,keyval in ipairs(tab) do
       if keyval.key == key then 
          return keyval.value 
       end
    end
    return 
end

---Clear user group cache.
--  @return a table with one entry; httpStatusCode.
--  @usage
--  local resp=rfw.rfw_flush_cache()
--  print(resp['httpStatusCode'])
function M.rfw_flush_cache()
   Trace("rfw_flush_cache: ")
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return 
   end
   M.user_workgroups = {}
   local response={}
   local hpass, hout_str, hcode  = rest.request( 'POST', "config/clearUserGroupCache", nil, "Accept:*/*; q=0.01",'application/json;charset=UTF-8')
   if not hpass then
      sas.print("%1z"..tostring(hout_str))
   else     
      sas.print("%3z"..tostring(hcode))
   end
   response.httpStatusCode=hcode   
   return response
end 

--- Escape special characters (from a json sense) in a string
-- @param 'in_str' A string intended to be included as part of a json string
-- @return 'result' The input string with the following characters escaped: \ " \r \n \t (last 3 are special whitespace characters).
function M.escape_json_string_value(in_str)
   Trace("escape_json_string_value: ",in_str)
   if in_str then
      local result = tostring(in_str)
      --Slash and double quote
      result = string.gsub(result, '([\\"])', '\\%1')
      --Carriage return
      result = string.gsub(result, '\r', '\\r')
      --Line feed
      result = string.gsub(result, '\n', '\\n')
      --Tab
      result = string.gsub(result, '\t', '\\t')
      return result
   else
      return ""
   end
end

--[[****************************************************
   *                    #WORKSHEET                     *
   *****************************************************]]--
   
--- Worksheet Template Functions
-- @section Worksheet

--- Create a worksheet template from a Microsoft Excel file. The file name for the input file must conform to the following syntax:
-- 
--  `*template-category*``_``*template-name*`.`*extension*`
-- 
--  * `*worksheet-category*`: the worksheet template category that the new template will be associated with, when the template
--    is created.
--  * `*template-name*`: the name to use for the worksheet template, when it is created.
--  * `*extension*`: the file extension. SAS Risk and Finance Workbench supports .XLS or .XLSX files.
-- 
--  For example: `Samples_Template1.xlsx`
-- 
--  Use the following columns in the Microsoft Excel file to define the worksheet template:
-- 
--  * Axis (required): whether the specified hierarchy is displayed on an axis or acts as a filter. Specify `X` or `Y` for an axis
--    or `Slicer` for a filter.
--  * Hierarchy_Name (required): the name of the hierarchy to include in the worksheet template.
--  * Hierarchy_Version (optional): the version of the hierarchy to include. The default value is `""`.
--  * History_Periods (required for time hierarchies): the number of history time periods to include in the worksheet, if you are 
--    specifying a time hierarchy. For other types of hierarchies, leave this field blank.
--  * Future_Periods (required for time hierarchies): the number of future time periods to include in the worksheet, if you are 
--    specifying a time hierarchy. For other types of hierarchies, leave this field blank.
--  * Version (optional): an identifier to distinguish this worksheet template from other worksheet templates with the same name. 
--    The default value is `""`.
--  * Description (optional): the description of the worksheet template. SAS Risk and Finance Workbench evaluates only the first row
--    in the input file for the Description field. Values in other rows are ignored.
--  * Members (optional): the subset of hierarchy members to include in the template, if you do not want to include the entire
--    hierarchy. To represent the refined hierarchy, separate the members with the `#` character (for example,
--    `Country#Region`). This field corresponds to the **Refine** feature in the user interface.
--  * Subtotals (optional): whether to display subtotals for rows in the rendered worksheet. Valid values are `NO_SUBTOTALS` 
--    and `YES_SUBTOTALS`. The default value is `NO_SUBTOTALS`. This field corresponds to the **Include subtotal columns 
--    in worksheet** check box in the Refine window of the user interface. 
--  * aggregation (optional): Only specify in one row - "true" or "false", whether or not to aggregate missing values. Default is true. 
--
--  **Note:** At a minimum, your worksheet template must include exactly one entity hierarchy, exactly one time hierarchy, 
--  and at least one other hierarchy. Filters are not required.
-- 
-- @param input_file the full path to the Microsoft Excel file that defines the worksheet template to be created.
-- @param update[opt] a Boolean value that indicates whether to update the worksheet template if it already exists. The default value
-- is false.
-- @return `pass`: a Boolean value that indicates whether the worksheet template was created. A value of true indicates success.
-- @return `out_str`: a message from the server with additional information about the operation.
-- @usage M.worksheet_create("/usr/tmp/Samples_Template1.xlsx", true)

function  M.worksheet_create(input_file, update)
   Debug("--- worksheet_create --- ")
   check_args( input_file, update,{["fn"]="worksheet_create", "string",{"boolean",'nil'}} )
   Debug("Loading template ",input_file)
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return false, nil
   end

   local hier_name, hier_ver, hier_key, hier_start_lev, hier_end_lev, hier_code, hierarchies, worksheet, data_exists,
      body, description, version, wrksht_ind, wrksht, hier_ind, hier, i, v, temp_name, temp_cat, file_name,
      file_sep, categoryKey, pass, nrows, is_time, fut_periods, hist_periods, out_str, code

   sas.submit([[
      
      PROC IMPORT OUT= WORK._temp_ 
                  DATAFILE= "@file@" 
                  DBMS=xlsx REPLACE;
           GETNAMES=YES;
      RUN; 
      
      /* Get the Order for each entry */
      data _temp_; set _temp_;
         if Axis eq "Slicer" then do;
            Sorter = "Z";
         end;
         else do;
            Sorter = Axis;
         end;
      run;
      proc sort data = _temp_; by Sorter; run;
      data _temp_; 
         set _temp_;
         by Sorter;
         retain Order 0;
         if first.Sorter then Order = 0;
         Order = Order + 1;
      run;
   ]], {file = input_file})
   data_exists = sas.exists("work._temp_")
   if not data_exists then
--       sas.print("%1z The Worksheet was not loaded.")
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_worksheet_create_1', 'NOQUOTE'))
      return false
   end
   worksheet = {}
   worksheet = sas.load_ds("work._temp_")
   body = '{"axes":['
   hierarchies = M.hierarchy_get()   
   local hier_details = M.hierarchy_get_details()
   
   --Need to know how many hierarchies are being used
   nrows = #worksheet
 
   -- Get the category and name from the file name. First get the file name
   -- without the path, then get category and name. Assumes cat_name convention
   file_sep = utils.get_file_sep()
   file_name = sas.scan(input_file,-1,file_sep)
   local index_1, index_2
   index_1 = sas.findc(file_name,'_')
   index_2 = sas.findc(file_name,'.','b')
   temp_cat = sas.substr(file_name,1,index_1-1)
   temp_name = sas.substr(file_name,index_1+1,index_2-index_1-1)
   Debug("Creating Worksheet ",temp_name," in category ",temp_cat)

   -- Loop over all the rows of the worksheet table. For each hierarchy specified,
   -- need to then search through the hierarchies to find the associated key and other details
   fut_periods = -1
   hist_periods = -1
   description = ""
   version = ""
   local aggregation = "true"
   for wrksht_ind,wrksht in ipairs(worksheet) do
      hier_name = tostring(wrksht.hierarchy_name)
      hier_ver  = tostring(wrksht.hierarchy_version or "")
      if wrksht.description and wrksht.description ~= "" then 
         description = tostring(wrksht.description)
      end
      if wrksht.version and wrksht.version ~= "" then 
         version = tostring(wrksht.version)
      end
      if wrksht.aggregation and wrksht.aggregation ~= "" then 
         aggregation = tostring(wrksht.aggregation)
      end
      
      -- If members are specified, need to figure out the start level and end level, so need the max and min level number for all specified.
      local li_key_map = {}
      if wrksht.members and wrksht.members ~= "" then
         for i,v in ipairs(hier_details) do
           v.hierarchy.version = v.hierarchy.version or ""
           if sas.lowcase(v.hierarchy.version) == sas.lowcase(hier_ver) and sas.lowcase(v.hierarchy.name) == sas.lowcase(hier_name) then 
              -- map an li_key to a level number
              for _,line_item in ipairs(v.lineItems) do
                 li_key_map[line_item.liKey] = line_item.levelNum
              end
              break
           end
         end
      end         
      -- Look for this hierarchy in the table and get the details needed
      hier_key = -1
      for hier_ind, hier in pairs(hierarchies) do
         if not hier.version then hier.version = '' end
         if sas.lowcase(hier.name) == sas.lowcase(hier_name) and sas.lowcase(hier.version) == sas.lowcase(hier_ver) then
            hier_key = hier.key
            if wrksht.members and wrksht.members ~= "" then
               local members_table = string.split(wrksht.members,'#')
               hier_start_lev = li_key_map[members_table[1]]
               hier_end_lev = 1
               for _,mem in ipairs(members_table) do
                  hier_start_lev = math.min(li_key_map[mem],hier_start_lev)
                  hier_end_lev = math.max(li_key_map[mem],hier_end_lev)
               end
            else
               hier_start_lev = 1
               hier_end_lev = math.max(#hier.levelNames,1)
            end
            hier_code = hier.code
            if hier_code == 'TIME' then
               is_time = true
               fut_periods = wrksht.future_periods
               hist_periods = wrksht.history_periods
            else 
               is_time = false
            end
            body = body .. '{"hierarchyKey":'..hier_key..',"code":"'..hier_code..'","axis":"'..wrksht.axis..'","order":'..wrksht.order..',"startLevel":'..hier_start_lev..',"endLevel":'..hier_end_lev
            --Handlde any Refinements
            body = body ..',"members":['
            if wrksht.members then
               for i,mem in ipairs(wrksht.members:split('#')) do
                  if i ~=1 then body = body..',' end
                  body = body..'"'..mem..'"'
               end
            end
            body = body..']'
            if sas.upcase(tostring(wrksht.subtotals))=='YES_SUBTOTALS' then
               if sas.upcase(tostring(wrksht.axis)) ~= 'X' then
                  sas.print('%2z Can not enable subtotals for hierarchy '..hier.name..', version '..hier.version..' because it is not on the X axis')
               else
                  body = body.. ',"subtotalCode":"YES_SUBTOTALS"'
               end
            end
            body = body .. ',"name":"'.. hier_name..'"}'
            if wrksht_ind ~= nrows then
               body = body .. ','
            end --Not on last row (last row doesn't get a comma)
            Debug("For hierarchy ",hier_name," have:")
            Debug("is_time = ",is_time)
            Debug("fut_periods = ",fut_periods)
            Debug("hist_periods = ",hist_periods)
            break            
         end --hierarchy name and version match
      end --Loop over hierarchies
      if hier_key == -1 then sas.print("%1z Hierarchy "..hier_name.." specified in template "..input_file.." was not found") return end
   end-- Loop over entries in worksheet table
   
   -- Find the category key, if it doesn't exist, make the category
   categoryKey, pass = M.worksheet_category_key_get(temp_cat)
   if not pass then
      -- If the categoryKey doesn't exist then create that category
      M.worksheet_category_create(temp_cat)
      categoryKey, pass = M.worksheet_category_key_get(temp_cat)
   end   
   body = body..'],"name":"' .. temp_name ..'","version":"'.. version ..'","description":"'.. description .. '","categoryKey":"' .. categoryKey ..'","numFuturePeriods":"'..fut_periods..'","numHistoryPeriods":"'..hist_periods..'","aggregateMissingValues":'..aggregation..'}'
   utils.write_fileref("in", body)
   local url_append = 'config/sheets/templates'
   
   -- Check to see if the worksheet exists already. If it does and update is true, process with request_type 'PUT', otherwise abort.
   local wrksht_key
   local request_type = 'POST'
   local pass, json_out, code = rest.request( 'GET', 'config/sheets/templates') 
   if not pass then 
      Debug("Unable to get the list of worksheets: ",json_out) 
      return false 
   end
   local sheets = Json:decode(json_out)   
   --Search for the worksheet with the supplied name. 
   for i,v in pairs(sheets["items"]) do
      if not v.version then v.version = '' end
      if sas.lowcase(v.name) == sas.lowcase(temp_name) and sas.lowcase(v.version) == sas.lowcase(version) then
         if update then
            request_type = "PUT"         
            wrksht_key = v["key"]
            url_append = url_append.."/"..tostring(wrksht_key)
            break
         else
            return false
         end
      end
   end
   pass, out_str, code = rest.request(request_type, url_append, {[1]='in'}, 'Accept:application/json \nAccept-Encoding:gzip, deflate', 'application/json;charset=UTF-8', 'out') 
   return pass, out_str
end

--- Create a worksheet template category
--  @param name the name of the category to create.
--  @return `pass`: a Boolean value that indicates whether the template category was created. A value of true indicates success.
--  @return `reason`: the message that is returned from the server in response to the HTTP request.
function M.worksheet_category_create(name)
   check_args( name,{["fn"]="worksheet_category_create", "string"})
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return false 
   end
   name = M.escape_json_string_value(name)
   sas.submit([[
      data _null_;
         file in;
         put '{"categoryKey":0,"name":"@name@","categoryCode":0,"readOnly":false,"hierarchiesCount":0}';
      run;
      data _null_;
         file hin;
         put 'Accept:application/json';
         put 'Accept-Encoding:gzip, deflate';
      run; 
   ]],{name=name})
   local pass, out_str, code = rest.request("POST", "config/sheets/categories", {[1]='in'}, {[1]='hin'}, 'application/json;charset=UTF-8', nil)  
   return pass, out_str
end

--- Retrieve the key value that identifies a worksheet template category.
--  @param name the name of the worksheet template category.
--  @return `categoryKey`: the numeric key value for the specified template category.
--  @return `pass`: a Boolean value that indicates whether the template category was found. A value of true indicates success.
function M.worksheet_category_key_get(name)
   check_args( name,{["fn"]="worksheet_category_key_get", "string"} )
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return nil, false 
   end
   local categoryKey = nil
   local categories = M.worksheet_category_get_all()
   for i,v in pairs(categories) do
      if categories[i].name == name then
         categoryKey = categories[i].categoryKey
         return categoryKey, true
      end
   end    
   return nil, false
end
        
--- Retrieve a table that contains all worksheet template categories.
--  @return `categories`: a table with the worksheet template categories. For each template category, the table includes the template category name
--  and additional information, such as the category key.
function M.worksheet_category_get_all()
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return nil
   end
   local categories
   local pass, out_str, code = rest.request("GET", "config/sheets/categories", nil, 'Accept:application/json, text/javascript, */*; q=0.01 \nAccept-Encoding:gzip, deflate, sdch')
   if pass then
      categories = Json:decode(out_str)
      categories = categories.items
   end
   return categories
end

--- Retrieve the key value that identifies a worksheet template.
--  @param name the name of the worksheet template.
--  @param version the version identifier that distinguishes this worksheet template from other worksheet templates with the same name.
--  The default value is `""`.
--  @return `key`: the numeric key value for the specified worksheet template. If the worksheet is not found, `nil` is returned.
function M.worksheet_key_get(name, version)
   check_args( name, version, {["fn"]="worksheet_key_get", "string",{"string",'nil'}} )
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return nil
   end
   version = version or ""
   --First get the worksheets as a table
   local pass, json_out, code = rest.request( 'GET', 'config/sheets/templates') 
   local key = nil
   local sheets = Json:decode(json_out)   
   --Search for the worksheet with the supplied name
   for i,v in pairs(sheets["items"]) do
      if sas.lowcase(v.name) == sas.lowcase(name) and sas.lowcase(v.version) == sas.lowcase(version) then
         key = v.key
         return key
      end
   end
   
   return nil
end

--- Save data in a worksheet to a sas dataset
--  @param scenario_names      - Table with list of scenario names. e.g. {"scen1","scen2"}. Output id for all scenarios listed. For no scenarios, pass nil
--  @param project_name_or_key - Name of the project or the project key
--  @param sheet_name          - The name of the worksheet or table with entries name and version
--  @param entity_name - The name of the entity to get data for
--  @param output_ds - The sas dataset to write to
--  @param other_filters - table, any other likeys to filter on, e.g. {likey_1,likey_2,likey_3}. Pass nil or an empty table if none.
--  @param remote[OPT] - Path of the public workgroup folder for rfw. If nil the root directory is pulled from the return body provided by the server. Passing a value for this is intended as a tool for testing. In practice it should be left nil.
--  @return pass - Boolean whether or not the function call succeeded
function M.worksheet_data_get(scenario_names, project_name_or_key, sheet_name, entity_name, output_ds, other_filters, remote )
   check_args( scenario_names, project_name_or_key, sheet_name, entity_name, output_ds, other_filters, remote,{["fn"]="worksheet_data_get", {"table",'nil'},{"string",'number'},{"string",'table'},{'string','nil'},'string',{'table','nil'},{'string','nil'}} )
   if not M.url_set(nil) then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return false 
   end
   
   other_filters = other_filters or {}
   local worksheet_version, worksheet_name, version_url_portion
   if type(sheet_name) == 'table' then
      worksheet_name = sheet_name.name
      worksheet_version = sheet_name.version
      version_url_portion = '&sheetVersion=' .. worksheet_version
   else
      worksheet_name = sheet_name
      version_url_portion = ''
   end
   
   -- generate a random name, 6 chars long to extract the sheet to
   local rand_name= "_"
   for i=1, 5 do
      rand_name =rand_name .. sas.byte(sas.int(65+26*sas.ranuni(0)))
   end
   Debug("Random name=", rand_name)
   
   -- Get the project key if necessary
   local project_key = project_name_or_key
      if type(project_name_or_key) == 'string' then
         project_key = M.project_key_get(  project_name_or_key)
         if project_key == nil then
            -- %%1zInvalid project name <%%s> or no projects match.
            sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_project_key_get_1', 'NOQUOTE'),project_name_or_key)
            return false
      end
   end

   -- Get the entity name likey
   local entity_key = M.entity_likey_get( entity_name, project_key)
   if entity_key == nil then
      -- The entity %%s doesn't exist in this project.
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_model_group_agg_load_setup_3', 'NOQUOTE'),entity_name)
      return false
   end
  
   -- Loop over the scenarios
   scenario_names = scenario_names or {""}
   Debug("Scenario names")
   Debug(scenario_names)
   local ignore_vars = {  measure_value = 1, measure_date=1,  period_type=1, scenario=1 }

   for scen_num,s in ipairs( scenario_names ) do
      -- Get scenario liKey
      local scenarioLiKey =""
      if s ~= "" then
         local liKey = M.scenario_likey_get(s,project_key)
         if liKey == nil then 
            -- %%1zThe scenario %%s doesn't exist in this project.
            sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_model_group_agg_load_setup_4', 'NOQUOTE'),s)
            return false 
         end
         scenarioLiKey = "&scenarioName=" .. s
      end
      local json, pass
      local extra_filters=""
      --Add extra filters
      for i,f in ipairs(other_filters) do
         extra_filters=extra_filters.."&filterLiKey="..tostring(f)
      end

      local request = "projects/" .. project_key .. "/sheets/byName?exportValueTypes=TN&sheetName=" .. worksheet_name .. version_url_portion .. '&outputFilename='.. rand_name .. '.xlsx'.. "&filterLiKey=" .. entity_key .. scenarioLiKey..extra_filters
      pass, json_out = rest.request( "post", request)

      if not pass then
         sas.print("%1z"..tostring(json_out))
         return false
      end

      local returned_table = Json:decode(json_out)
      local path_to_excel = remote or returned_table.serverFilepathToExportedSheet
   
      sas.submit([[
         proc import out=sheet_data
            datafile="@path_to_excel@"
            dbms=xlsx replace;
         run;  
      ]])

      -- Delete the exported data excel file
      sas.filename("__temp_", path_to_excel);
      sas.fdelete("__temp_")
      sas.filename("__temp_")

      if scen_num == 1 then
         sas.submit([[ data @output_ds@; set sheet_data; run; ]])
      else
         sas.submit([[ proc append base=@output_ds@ data=work.__tmp force;run;]])
      end
   end
   return true
end
   
--[[****************************************************
   *                      #DATA                        *
   *************************************************]]--

--- Data Loading Functions
-- @section Data

--- Load global data from the staging directory on the SAS Risk and Finance Workbench server. You can retrieve the location of the staging 
--  directory with the `staging_dir_get` function.
-- 
--  **Note:** This is deprecated - you should use the generic 'data_load' function. 
function M.data_global_load(file_name, dataFormat, dataLoadMapName, workgroup, dataSource)
   return M.data_load(file_name, nil, dataFormat, dataLoadMapName, workgroup, dataSource)
end 
   
--- Load project data from the staging directory on the SAS Risk and Finance Workbench server. You can retrieve the location of the staging 
--  directory with the `staging_dir_get` function.
-- 
--  **Note:** This is deprecated - you should use the generic 'data_load' function. 
function M.data_project_load(file_name, project, dataFormat, dataLoadMapName, workgroup, dataSource) 
   return M.data_load(file_name, project, dataFormat, dataLoadMapName, workgroup, dataSource)
end

--- Load data. This is an all-purpose data loading function meant to replace the two previous functions data_project_load and data_global_load. To load global
--- data, pass nil as the project
-- 
--  @param file_name: string, The name of the file to load, including the file extension. If the file is in a subdirectory within the staging area, specify the path to
--                    the file relative to the staging area.So, if the file is in admin workgroup, in a subfolder called sub1. Pass 'sub1/file_name.ext' and pass 
--                    the value 'admin' as the workgroup argument.
--  @param project[OPTIONAL]: string or number, If string, then this is the name of the project where you want to load data. If you pass a number, then this is the project key. 
--                            Passing the key directly is more efficient as otherwise an extra rest call is made to look it up. If nil, then data is loaded as global data.
--  @param dataFormat[OPTIONAL]: string, the type of  data format. Must be one of: "HierarchyBased", "DimensionBased", "Mapped". Default is "HierarchyBased"
--  @param dataLoadMapName[OPTIONAL]: string, The name of the data load map to use. Necessary if dataFormat == "Mapped"
--  @param workgroup[OPTIONAL]: string, the name of the workgroup where the file to load is located. Defaults to 'public'.
--  @param dataSource[OPTIONAL]: string, D or O. This specifies whether the data are to be labeled as coming from a standard data load (D)
--                               or as an external override (O). Default is D.
--  @return 'msg': string, the output body from the HTTP request.
--  @return 'pass': boolean, true if load was succesful, false otherwise.
function M.data_load(file_name, project, dataFormat, dataLoadMapName, workgroup, dataSource)
   Trace("data_project_load",file_name, project, dataFormat, dataLoadMapName, workgroup, dataSource)
   check_args( file_name, project, dataFormat, dataLoadMapName,workgroup,dataSource,{["fn"]="data_project_load", "string",{'string','nil','number'},{'string','nil'},{'string','nil'},{'string','nil'},{'string','nil'}} )
   
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return nil, false 
   end

   --defaults
   workgroup = workgroup or 'public'
   dataFormat = dataFormat or 'hierarchybased'
   dataSource = dataSource or 'D'

   -- Check the data_format arg
   dataFormat = dataFormat:lower()
   local valid_formats = {hierarchybased=true,dimensionbased=true,mapped=true}
   if not valid_formats[dataFormat] then
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_data_load_bad_format', 'NOQUOTE'))
      return sas.sasmsg('sashelp.stressluamsg',  'sst_data_load_bad_format', 'NOQUOTE'), false
   end
   
   -- Check the dataSource arg
   dataSource = dataSource:upper()
   valid_formats = {D=true,O=true}
   if not valid_formats[dataSource] then
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_data_load_bad_source', 'NOQUOTE'))
      return sas.sasmsg('sashelp.stressluamsg',  'sst_data_load_bad_source', 'NOQUOTE'), false
   end
   
   --Construct the surrogateFilePath paramater
   local surrogateFilePath = workgroup..'/'..file_name
   
   -- Get the project key if project data
   local key
   if project then
      if type(project) == 'string' then
         key = M.project_key_get(project)
         if not key then
            sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_project_key_get_1', 'NOQUOTE'),project)
            return sas.sasmsg('sashelp.stressluamsg',  'sst_project_key_get_1', 'NOQUOTE'),false
         end
      else
         key = project
      end
   end
   
   -- Get the base url for the request
   local url
   if project then
      url = "projects/"..key.."/projectDataFromServer"
   else
      url = "config/globalDataFromServer"
   end
   
   -- Put together any parameters to pass in the url
   local function add_parm_to_url(url,parm)
      if url:find('?') then
         return url..'&'..parm
      else
         return url..'?'..parm
      end
   end
   
   if dataFormat~= "hierarchybased" then
      url = add_parm_to_url(url,"dataFormat="..dataFormat)
   end
   if dataFormat == "mapped" then
      url = add_parm_to_url(url,"dataLoadMapName="..utils.urlencode(dataLoadMapName))
   end
   url = add_parm_to_url(url,"dataPointSource="..utils.urlencode(dataSource))
   url = add_parm_to_url(url,"surrogateFilePath="..utils.urlencode(surrogateFilePath))
             
   local pass, json_out, code = rest.request( 'POST', 
                                              url,
                                              'Accept: */* \nAccept-Encoding:gzip, deflate'
                                              )
   return json_out, pass
end

--- Create a Data Load Map from a file
--
-- @param name - string, the name of the Data Load Map to create
-- @param description[OPTIONAL] - string, The description for the Data Load Map
-- @param file - string, the fully qualified path to the file that will be loaded as a Data Load Map
-- @param update[OPTIONAL] - boolean, true if you want to update an existing map with the same name. Defaults to false.
-- @return pass - boolean, whether or not the call was successful. True is success.
-- @return msg - string, output body of the rest request
function M.data_load_map_create(name, description, file, update)
   Trace("data_load_map_create",name, description, file, update)
   check_args( name, description, file, update,{["fn"]="data_load_map_create", "string",{'string','nil'},'string',{'boolean','nil'}} )
   
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return nil, false 
   end
   
   local file_name = sas.scan(file,-1,'\\/')
   
   --Defaults
   description = description or ""
   update = update or false
   
   -- See if the map already exists
   local dlm_key = M.data_load_map_key_get(name)
   if dlm_key and not update then
      return false, sas.sasmsg('sashelp.stressluamsg',  'sst_data_map_already_exists', 'NOQUOTE')
   end
   local url = 'dataLoadMaps'
   if dlm_key then
      url = url..'/vars/'..tostring(dlm_key)
   end
   
   -- Create the body for the request
   local body = '------WebKitFormBoundaryWtVKg8WEiVgBo3F8\r\n'
   body = body.. 'Content-Disposition: form-data; name="file"; filename="'..file_name..'"\r\n'
   body = body.. 'Content-Type: application/vnd.openxmlformats-officedocument.spreadsheetml.sheet\r\n\r\n'
   
   local mpdta_logical = sasxx.assign(file,"mpdta")
   local handle = mpdta_logical:open('r-o','fc')
   if not handle then
--       sas.print("%1z Could not open "..file..". Check that it exists and is readable")
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_process_create_1', 'NOQUOTE'),file)
   end
   for line in handle:lines() do
      body=body..line
   end
   handle:close()
   mpdta_logical:deassign()
   body = body.. '\r\n'
   body = body.. '------WebKitFormBoundaryWtVKg8WEiVgBo3F8\r\n'
   body = body.. 'Content-Disposition: form-data; name="_charset_"\r\n\r\n'
   body = body.. 'UTF-8\r\n'
   body = body.. '------WebKitFormBoundaryWtVKg8WEiVgBo3F8\r\n'
   body = body.. 'Content-Disposition: form-data; name="file-data"\r\n\r\n\r\n'
   body = body.. '------WebKitFormBoundaryWtVKg8WEiVgBo3F8\r\n'
   body = body.. 'Content-Disposition: form-data; name="name"\r\n\r\n'
   body = body.. name..'\r\n'
   body = body.. '------WebKitFormBoundaryWtVKg8WEiVgBo3F8\r\n'
   body = body.. 'Content-Disposition: form-data; name="description"\r\n\r\n'
   body = body.. description..'\r\n'
   body = body.. '------WebKitFormBoundaryWtVKg8WEiVgBo3F8--'  

   local pass, json_out = rest.request( 'POST', 
                                      url,
                                      body, 
                                      'Accept: */* \nAccept-Encoding:gzip, deflate', 
                                      'multipart/form-data; boundary=----WebKitFormBoundaryWtVKg8WEiVgBo3F8'
                                      )  

   return pass,json_out
end

--- Get a table with all Data Load Maps
-- @return maps - table, all data load maps. nil if error.
-- @return msg - string, output body of the rest request (if successful, returns output header since body is table with all maps)
function M.data_load_map_get_all()  
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return nil, false 
   end
   local pass, out = rest.request('GET','dataLoadMaps')
   if not pass then 
      return nil, out 
   end
   return Json:decode(out).items, utils.read_fileref('hout')
end

--- Get the key associated with a data load map
--
-- @param name - string, the name of the Data Load Map to get the key for
-- @return key - number, key for the data load map. nil if error or map does not exist.
-- @return msg - string, output body of the rest request 
function M.data_load_map_key_get(name)
   Trace("data_load_map_create",name)
   check_args( name,{["fn"]="data_load_map_create", "string"} )
   
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return nil, false 
   end
   
   local maps, out = M.data_load_map_get_all()
   if not maps then 
      return nil, out
   end
   local key
   for _,m in ipairs(maps) do
      if m.name:lower() == name:lower() then
         key = m.key
         break
      end
   end
   return key, out
end

--- Delete a Data Load Map
--
-- @param name - string, the name of the Data Load Map to delete
-- @return pass - boolean, whether or not the call was successful. True is success.
-- @return msg - string, output body of the rest request 
function M.data_load_map_delete(name)
   Trace("data_load_map_create",name)
   check_args( name,{["fn"]="data_load_map_create", "string"} )
   
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return nil, false 
   end
   local key, msg = M.data_load_map_key_get(name)
   if not key then
      return false,msg
   end
   local pass, msg = rest.request('DELETE','dataLoadMaps/'..tostring(key))
end


-- *****************************************************
-- *                    #PROJECT                       *
-- *****************************************************

--- Project Functions
-- @section Projects

--- Create a new project.  
--  @param name the name of the project to create.
--  @param description the description for the new project.
--  @param process the name of the process definition to use with the project. You can specify `nil` if you do not want to use a 
--  process definition with the project.
--  @param as_of_date the as-of date for the project. Specify the date as a string with the format `YYYY-MM-DD`.
--  @param notifications whether to send email notifications to the participants in the project workflow.
--  Specify `1` to send notifications. The default value is `0` (no notifications). If you prefer, you can 
--  specify Boolean values instead.
--  @param Worksheets[opt] an integer-indexed table with the names of worksheet templates to include in the project. When you define
--  the table, you can specify template names with versions or only template names. For example:
-- 
--  **Name and version:** `{[1]={name = "Worksheet1", version = "v1"}}`
--
--  **Name only:** `{[1] = "WorksheetA", [2] = "WorksheetB"}` 
-- 
--  When you specify names only, the versions default to `nil`.
--  @return `pass`: a Boolean value that indicates whether the project was created. A value of true indicates success.
function M.project_create(name, description, process, as_of_date, notifications, Worksheets)
   check_args( name, description, process, as_of_date, notifications, Worksheets, 
{["fn"]="project_create", "string","string",{"string",'nil'},"string",{"string",'nil','boolean'},{"table",'nil'}} )
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return nil, false 
   end
   -- notifications defaults to none
   if not notifications then notifications = '0' else notifications = '1' end
   local prc_sds, key  
   -- Set the processStartDateStr. and process key. If no process the date is null and key is 0
   -- otherwise date is today's date and key needs to be passed
   if not process then
      prc_sds = 'null'
      key = 0
   else
      local dt = sas.today()
      local yr = sas.year(dt)
      local mnth = sas.month(dt)
      local dy = sas.day(dt)   
      prc_sds = '"'..yr..'-'..mnth..'-'..dy..'"'
      key = M.process_key_get(process)
      if not key then
   --          sas.print("%1z Unable to find the process " ..process)
            sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_project_create_1', 'NOQUOTE'),process)
         return false
      end
   end
   
   local body = '{"projectName":"'..name..'","description":"'..description..'","processDefinitionKey":'..key..',"asOfDateStr":"'..as_of_date..'","processStartDateStr":'..prc_sds..',"processNotificationType":'..notifications
   
   -- If we are attaching worksheets to the project, get their keys and add them to the request
   local templateKey
   if Worksheets then
      body = body..',"sheetDefinitions":['
      for i,v in ipairs(Worksheets) do
         local templateKey
         if type(v) == 'table' then
            templateKey = M.worksheet_key_get(v.name,v.version)
         else
            templateKey = M.worksheet_key_get(v)
         end
         if templateKey then
            if i > 1 then
               body = body..','
            end
            body = body..'{"sheetTemplateKey":'..templateKey..'}'
         else
--             sas.print("%1z Was unable to find a worksheet named "..v)
            sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_project_create_2', 'NOQUOTE'),v)
         end
      end
      body = body..']'
   end
   body = body..'}'
   sas.submit([[
      data _null_;
         file _hin_;
         put 'Accept:application/json';
         put 'Accept-Encoding:gzip, deflate';
      run;  
      ]])
    local pass, json_out, code = rest.request( 'POST', 'config/projects/', body, {[1] = '_hin_'}, 'application/json;charset=UTF-8', nil) 
    return pass

end

--- Delete a project.  
--  @param name the name of the project to delete.
--  @return `pass`: a Boolean value that indicates whether the project was deleted. A value of true indicates success.
--  @return `msg`: a message from the server with additional information about the operation.
function M.project_delete(name)
   check_args( name, {["fn"]="project_delete", "string"} )
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return false, nil 
   end
   local key = M.project_key_get(name)
   if not key then
      return false, nil
   end
   local pass, json_out, code = rest.request( 'DELETE', 'config/projects/','['..key..']' , 'Accept:application/json \nAccept-Encoding:gzip, deflate, sdch', 'application/json', nil) 
   return pass, utils.read_fileref('out')
end
   
--- Retrieve a list of all projects on the SAS Risk and Finance Workbench server. 
--  @return `projects`: a table of the projects on the server.
--  @return `pass`: a Boolean value that indicates whether the list of projects was retrieved. A value of true indicates success.
function M.project_get( )
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return nil, false 
   end
   local pass, json_out, code
   local projects = {}

   pass, json_out, code = rest.request("GET", "projects", nil, "Accept: */* \nAccept-Encoding:gzip, deflate", "application/x-www-form-urlencoded; charset=UTF-8", nil)  
   if not pass then 
      sas.print('%2z'..json_out)
   else
      projects = Json:decode(json_out) 
   end
   return projects, pass
end

--- Retrieve the current iteration for a specified project.
--  @param name the name of the project.
--  @return `iteration`: the current iteration for the specified project. If the project is not found, `nil` is returned.
--  @return `pass`: a Boolean value that indicates whether the iteration was retrieved. A value of true indicates success.
function M.project_iteration_get(name)
   check_args( name, {["fn"]="project_iteration_get", "string"} )
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return nil, false 
   end

   local iteration = nil
   local pass = false
   local projects, pass = M.project_get()
   if pass then
      for i,v in pairs(projects["items"]) do
         if projects["items"][i]["name"] == name then
            iteration = projects["items"][i]["currentIterationNum"]
            return iteration, true
         end
      end 
   end
   return nil, false
end
        
--- Retrieve the key value that identifies a project.
--  @param name the name of the project.
--  @return `key`: the key value for the project. If the project cannot be found or the key cannot be retrieved, `nil` is returned.
function M.project_key_get(name)
   check_args( name, {["fn"]="project_key_get", "string"} )
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return nil
   end
   if not M.url_set() then  return nil end
   
   local key = nil
   local projects, pass = M.project_get()
   if projects then
      for i,v in pairs(projects.items) do
         if v.name == name then
            key = v.key
            break
         end
      end
   end
   return key
end
   
-- *****************************************************
-- *                    #PROCESS                       *
-- *****************************************************

--- Process Definition Functions
-- @section Process

--- Create a process definition from a Microsoft Excel file. 
--  @param file the full path to the process definition file. 
--  @return `out_str`: the output body from the HTTP request.
--  @return `pass`: a Boolean value that indicates whether the process definition was created. A value of true indicates success.
function M.process_create(file)
   check_args( file,{["fn"]="process_create", "string"})
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return nil, false 
   end
   local file_sep = utils.get_file_sep()
   local file_name = sas.scan(file,-1,file_sep)
  
   --BODY:
   local topText =    '------WebKitFormBoundarynL7rIPgAavvpQxNG\r\n'
   topText = topText.. 'Content-Disposition: form-data; name="file"; filename="'..file_name..'"\r\n'
   topText = topText.. 'Content-Type: application/vnd.openxmlformats-officedocument.spreadsheetml.sheet\r\n\r\n'

   local bottomText =    '\r\n------WebKitFormBoundarynL7rIPgAavvpQxNG\r\n'
   bottomText = bottomText.. 'Content-Disposition: form-data; name="name"\r\n\r\n'
   bottomText = bottomText.. file_name..'\r\n'
   bottomText = bottomText.. '------WebKitFormBoundarynL7rIPgAavvpQxNG--'

   local xl_logical = sasxx.assign(file,"xl")
   local in_logical = sasxx.new("in")
   
   local body = topText
   local handle = xl_logical:open('r-o','fc')
   if not handle then
--       sas.print("%1z Could not open "..file..". Check that it exists and is readable")
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_process_create_1', 'NOQUOTE'),file)
   end   
   for line in handle:lines() do
      body=body..line
   end
   handle:close()
   body = body..bottomText
   handle = in_logical:open('w+o',nil,string.len(body))
   handle:write(body)
   handle:close()

   xl_logical:deassign()   
   
   local pass, out_str, code = rest.request("POST", "tasks/definitions", {[1]='in'}, 'Accept-Encoding:gzip, deflate', 'multipart/form-data; boundary=----WebKitFormBoundarynL7rIPgAavvpQxNG', nil)    
   return out_str,pass
end

--- Retrieve a list of all process definitions on the SAS Risk and Finance Workbench server.
--  @return `processes`: a table with a list of the process definitions.
function M.process_get()
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return nil
   end
   Debug("--- process_get --- ")

   local pass, process_str, code = rest.request("GET", "tasks/definitions")    
   local json = Json:decode(process_str)
   local processes = json["items"]
   return processes
   
end

--- Retrieve the key value that identifies a process definition.
--  @param name the name of the process definition.
--  @return `key`: the value for the process definition. If the process definition cannot be found or the key cannot be retrieved, 
--  `nil` is returned. 
function M.process_key_get(name)
   check_args( name,{["fn"]="process_key_get", "string"} )
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return nil
   end
   Debug("--- process_key_get --- ")
   local processes = M.process_get()
   local key = nil

   for i,v in pairs(processes) do
      if string.trim(v["name"]) == string.trim(name) then
         key = v["key"]
         return key
      end
   end
   return nil
end

--- Retrieve the ID for the active workflow in a specified project. This ID identifies the workflow instance in the project and is not 
--  the ID for the process definition that was used to create the workflow.
--  @param project_id the numeric identifier for the project that contains the workflow. If you have only the project name, you can derive
--  the project ID by using the `project_key_get` function.
--  @return `process_key`: the numeric identifier for the workflow instance. This value enables requests to `/tasks/processes/*process_key*`.
function M.project_process_key_get( project_id )
   Trace("project_process_key_get", project_id)
   check_args(project_id,{["fn"]="project_process_key_get", "number"})
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return
   end
   -- First, get the project information so we can get the process definition key for the project
   local projects, pass = M.project_get( )
   if not pass then
      sas.print("%1zUnable to get the list of projects")
      return
   end
   local project
   for _,p in ipairs(projects.items) do
      if p.key == project_id then 
         project = p 
         break
      end
   end
   if not project then
--       sas.print("%1z Unable to find the project associated with project key")
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_task_complete_1', 'NOQUOTE'))
      return
   end
   return project.processKey
end

--- Delete a process definition. You cannot delete a process definition that is currently in use by a project.
-- 
--  **Note:** Because the object cannot be recovered, use caution when you delete a process definition.
--  @param Name the name of the process definition to delete.
--  @return `pass`: a Boolean value that indicates whether the process definition was deleted. A value of true indicates success.
--  @return `msg`: a message from the server with additional information about the operation.
function M.process_delete(Name)
   check_args(Name, {["fn"]="process_delete", "string"})
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return false,nil
   end
   local id = M.process_key_get(Name)
   if not id then return true,nil end
   local pass = rest.request( "DELETE", "tasks/definitions/"..tostring(id),nil,"Accept:*/*\nAccept-Encoding: gzip,deflate")
   return pass,utils.read_fileref('out')
end

-- *****************************************************
-- *                   #DIMENSION                      *
-- *****************************************************
--- Get a table with all dimension names and keys
-- 
--  @return dimensions: an indexed table - for each dimension have access to: createdByUser,createdTime,type,categoryKey,name,id,updatedByUser,levels,updatedTime,levelNames. 
function M.dimension_get_all()
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return nil
   end
   local pass, json_out = rest.request( "GET", "config/dimensions")
   return Json:decode(json_out).items
end

--- Get the id associated with a dimension
-- 
--  @param Name The name of the dimension to get the id for.
--  @return id. nil if dimension is not found
function M.dimension_id_get(name)
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return nil
   end
   local dims = M.dimension_get_all()
   for _,d in ipairs(dims) do
      if d.name == name then
         return d.id
      end
   end
   return
end

--- Get the id associated with a dimension category
--
-- @param name = the name of the dimension category
-- @return id nil if the category does not exist
function M.dimension_category_id_get(name)
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return nil
   end
   local pass, json_out = rest.request( "GET", "config/dimensions/categories")
   local cats = Json:decode(json_out).items
   for _,c in ipairs(cats) do
      if c.name:lower() == name:lower() then
         return c.id
      end
   end
   return
end

--- Create a dimension category
-- 
--  @param Name The name of the category to create
--  @return pass boolean whether or not the  dimension category was created.
function M.dimension_category_create(name)
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return nil
   end
   local id = M.dimension_category_id_get(name)
   if not id then
      Debug('Creating dimension category',name)
      local pass, out = rest.request('POST','config/dimensions/categories','{"categoryKey":0,"name":"'..tostring(name)..'","dimensionsCount":0}','Accept: application/json\nAccept-Encoding: gzip, deflate','application/json;charset=utf-8')
      Debug(out)
      return pass
   end
   return true
end

--- Create a dimension
-- 
--  @param Name The name of the dimension to create.
--  @param category The category to create the dimension in. It is created if it does not exist.
--  @param dim_type Acceptable values are: Account, Custom, Entity, Time
--  @param description[opt] Name The name of the dimension to create.
--  @param update[opt] Defaults to false. Boolean - whether or not to update it if it exists.
--  @return pass boolean whether or not the  dimension was created
--  @return out String containing any error message returned from the server
function M.dimension_create(name, category, dim_type, description,update)
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return nil
   end
   description = M.escape_json_string_value(description)
   update = update or false
   local dim_id = M.dimension_id_get(name)
   local request_type = 'POST'
   if update and dim_id then
      request_type = 'PUT'
   end
   if update or not dim_id then
      local pass = M.dimension_category_create(category)
      if not pass then return false end
      local cat_id = M.dimension_category_id_get(category)
      dim_type = dim_type:upper()
      dim_id = dim_id or 0
      local body = '{"id":"@dim_id@","categoryKey":"@cat_id@","name":"@name@","description":"@description@","type":"@dim_type@"}'
      local request_url = 'config/dimensions'
      if dim_id > 0 then
         request_url = request_url ..'/'..tostring(dim_id)
      end
      local pass, out = rest.request(request_type,request_url,body:resolve(),'Accept: application/json\nAccept-Encoding: gzip, deflate','application/json;charset=UTF-8')
      return pass, out
   else
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_dimension_already_exists', 'NOQUOTE'),name)
      return false, ""
   end
end

--- Delete a dimension
-- 
--  @param Name The name of the dimension to delete.f it exists.
--  @return pass boolean whether or not the  dimension was deleted
--  @return out String containing any error message returned from the server
function M.dimension_delete(name)
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return nil
   end
   local dim_id = M.dimension_id_get(name)
   if not dim_id then return true end
   local pass, out = rest.request('DELETE','config/dimensions/'..tostring(dim_id),nil,'Accept: */*')
   return pass, out
end



-- *****************************************************
-- *                   #HIERARCHY                      *
-- *****************************************************

--- Hierarchy Functions
-- @section Hierarchy


--- Create a hierarchy from a Microsoft Excel file. 
-- 
--  **Note**: The hierarchy cannot be created if any of the line items already exist in the system.
--  @param category the name of the hierarchy category to associate the hierarchy with. If the category does not exist, 
--  it is automatically created.
--  @param name the name of the hierarchy to create.
--  @param description the description to use for the hierarchy.
--  @param file the full path to the Microsoft Excel file that defines the hierarchy.
--  @param version[opt] the version for the hierarchy. If no version is specified, `""` is used.
--  @param update[opt] a Boolean value that indicates whether to update the hierarchy line items,
--  @param dimension[opt] The name of the dimension to use. If missing, the name of the hierarchy is used. If the dimension does not exist, it is created with type depending on the hierarchy category - Entity if entity, time if time, otherwise custom.
--  if the hierarchy already exists. The default value is true.                                                                                                                          
--  @return `pass`: a Boolean value that indicates whether the hierarchy was created. A value of true indicates success.
--  @return `out_str`: a message from the server with additional information about the operation.
--  @return `code`: the HTTP return code for the last request.
function M.hierarchy_create(category, name, description, file, version, update, dimension )
check_args( category, name, description, file, version, 
{["fn"]="hierarchy_create", "string","string","string","string",{"string",'nil'},{"string",'nil'}} )
   version = version or ""
   if update == nil then update = true end
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return false, nil, nil 
   end
   Debug("Calling hierarchy_create with args:",category, name, description, file, version )
   version = version or ""
   local file_sep = utils.get_file_sep()
   local file_name = sas.scan(file,-1,file_sep)

   -- Check if the hierarchy already exists.
   local ID, exists = M.hierarchy_key_get(name, version)
   Debug("Checking if hierarchy",name,version,"exists: ",exists,ID,". Update is set to: ",update)
   if exists and not update then
--       sas.print("%3z The hierarchy " .. name .. " with version "..version.." already exists and update is set to false.")
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_hierarchy_not_update', 'NOQUOTE'),name,version)
      return
   end
   
   -- First find the category key
   local categoryKey, pass = M.hierarchy_category_key_get(category)
   if not pass then
      -- If the categoryKey doesn't exist then create that category
--       sas.print("%3z Creating the hierarchy category " .. category)
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_hierarchy_create_1', 'NOQUOTE'),category)
      local hcc_pass, hcc_err = M.hierarchy_category_create(category)
      if not hcc_pass then
         return false,hcc_pass,nil
      end
      categoryKey, pass = M.hierarchy_category_key_get(category)
      if not pass then 
--          sas.print("%1z Unable to create the hierarchy category "..category) 
         sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_hierarchy_create_2', 'NOQUOTE' ),category)
         return
      end   
   end
   
   -- Creating a Hierarchy is done in 2 steps (2 http requests). First, the Hierarchy is created, 
   -- then the line items are loaded. If the hierarchy already exists, we can skip step 1.
   if not exists then
      local body_text = '{"hierarchyKey":0,"categoryKey":"@categoryKey@","categoryName":"@category@","name":"@name@","version":"@version@","description":"@description@"'
      -- If you don't specify a dimension, the server will try to figure out if it is Time, Entity, or Custom type and handle things. Otherwise, we specify in the request.
      local dimension_key
      if dimension then
         dimension_key = M.dimension_id_get(dimension)
         -- If a dimension was specified but it doesn;t exist, need to create it.
         if not dimension_key then
            local dim_type = 'CUSTOM'
            if category:upper() == 'TIME' then
               dim_type = 'TIME'
            elseif category:upper() == 'ENTITY' then
               dim_type = 'ENTITY'
            end
            local pass, out = M.dimension_create(dimension, dim_type, dim_type)
            if not pass then return pass, out end
            dimension_key = M.dimension_id_get(dimension)
         end
         body_text = body_text..',"dimensionName":"@dimension@","dimensionKey":"@dimension_key@"'
      end
      body_text = body_text..'}'
      body_text = string.resolve(body_text)
      local header_text = 'Accept:application/json\nAccept-Encoding:gzip, deflate'

      local  pass, out_str, code = rest.request("POST", "config/hierarchies", body_text, header_text, 'application/json;charset=UTF-8')
      if not pass then
         return pass, out_str, code
      end
   
      -- Now The hierarchy exists but the line items still need to be loaded. Get the ID
      -- associated with the hierarchy and then load line items.
      ID, exists = M.hierarchy_key_get(name, version)
      if not exists then
         sas.print("%1z Could not get the hierarchy key for hierarchy: "..tostring(name).." / "..tostring(version))
         return false, nil, nil
      end      
   end

   --Create the input header and body
   -- HEADER: 
   utils.write_fileref('hin','Accept-Encoding:gzip, deflate')

   --BODY:
   local topText =    '------WebKitFormBoundarynL7rIPgAavvpQxNG\r\n'
   topText = topText.. 'Content-Disposition: form-data; name="file"; filename="'..file_name..'"\r\n'
   topText = topText.. 'Content-Type: application/vnd.openxmlformats-officedocument.spreadsheetml.sheet\r\n\r\n'

   local bottomText =    '\r\n------WebKitFormBoundarynL7rIPgAavvpQxNG\r\n'
   bottomText = bottomText.. 'Content-Disposition: form-data; name="_charset_"\r\n\r\n'
   bottomText = bottomText.. 'UTF-8\r\n'
   bottomText = bottomText.. '------WebKitFormBoundarynL7rIPgAavvpQxNG\r\n'
   bottomText = bottomText.. 'Content-Disposition: form-data; name="file-data"\r\n\r\n\r\n'
   bottomText = bottomText.. '------WebKitFormBoundarynL7rIPgAavvpQxNG--'

   local xl_logical = sasxx.assign(file,"xl")
   local in_logical = sasxx.new("in")
    
   local body = topText
   local handle = xl_logical:open('r-o','fc')
   if not handle then
    --   sas.print("%1z Could not open "..file..". Check that it exists and is readable")
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_hierarchy_create_3', 'NOQUOTE'),file )
   end
   for line in handle:lines() do
      body=body..line
   end
   handle:close()
   body = body..bottomText

   handle = in_logical:open('w+o',nil,string.len(body))
   handle:write(body)
   handle:close()   
   xl_logical:deassign()   
   
   local pass, out_str, code = rest.request('POST', "config/hierarchies/"..ID.."/items", {[1]='in'}, {[1]='hin'}, 'multipart/form-data; boundary=----WebKitFormBoundarynL7rIPgAavvpQxNG','out')
   return pass, out_str, code
   
end

--- Retrieve the ID that is associated with a hierarchy.
--  @param name the name of the hierarchy.
--  @param version the version for the hierarchy. If no version is specified, `""` is used.
--  @return `ID`: the numeric identifier for the hierarchy. 
--  @return `pass`: a Boolean value that indicates whether the hierarchy ID was retrieved. A value of true indicates success.
function M.hierarchy_key_get( name, version )
   check_args( name, version, {["fn"]="hierarchy_key_get", "string", {"string",'nil'}} )
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return nil, false 
   end
   
   if not M.url_set() then  return false end
   version = version or ''
   
   local i, v
   local ID = nil
   local pass = false
   local hierarchies = M.hierarchy_get()

   if not hierarchies then
--       sas.print("%1zUnable to retrieve the hierarchies." )
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_hierarchy_key_get_1', 'NOQUOTE' ))
      return nil, false
   end

   for i,v in pairs(hierarchies) do
      if sas.lowcase(v.name) == sas.lowcase(name) and sas.lowcase(v.version or "") == sas.lowcase(version) then
         ID = v.id
         pass = true
         break
      end
   end
   Debug("Hierarchy ",name," has ID ",ID)
   return ID, pass
end

--- Retrieve a list of all hierarchies on the SAS Risk and Finance Workbench server.
-- @return `hierarchies`: a table with a list of all the hierarchies with associated information.
function M.hierarchy_get()
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return nil
   end
   local pass, json, code = rest.request( "GET", "config/hierarchies/", nil, "Accept:text/plain, */*; q=0.01")
   if not pass then
      local msg = rest.http_codes[code]
     -- sas.print("%1z GET request to config/hierarchies returned: "..tostring(code).." "..msg)
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_hierarchy_get_1', 'NOQUOTE'),tostring(code),msg)
      return nil
   else
      local out_string = utils.read_fileref("out")
      local hierarchies = Json:decode(out_string)
      hierarchies = hierarchies.items     
      return hierarchies
   end
end

--- Retrieve a list of all hierarchies with level information on the SAS Risk and Finance Workbench server.
-- @return `hier_details`: a table with a list of all the hierarchies with associated levelinformation.
function M.hierarchy_get_details()
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return nil 
   end
   local pass, code = rest.request( "GET", "config/hierarchies/details", nil, "Accept:text/plain, */*; q=0.01")
   if not pass then
      sas.print("%1z"..tostring(utils.read_fileref("hout")))
      return
   end  
   local hier_details = Json:decode(utils.read_fileref("out"))  
   return hier_details.items
end

--- Create a category for hierarchies.
--  @param name the name of the category to create.
--  @return `pass`: a Boolean value that indicates whether the hierarchy category was created. A value of true indicates success.
--  @return `out_str`: an informative message that is returned from the HTTP request.
function M.hierarchy_category_create(name)
   check_args( name, {["fn"]="hierarchy_category_create", "string"} )
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return false, nil 
   end
   local body_txt = '{"categoryKey":0,"name":"'..name..'","categoryCode":0,"readOnly":false,"hierarchiesCount":0}'
   local header_txt = 'Accept:application/json\nAccept-Encoding:gzip, deflate'
   local content_type = 'application/json;charset=UTF-8'
   local pass, out_str, code = rest.request("POST", "config/hierarchies/categories", body_txt, header_txt, content_type, nil)
   return pass, out_str
end

--- Retrieve the key value that is associated with a hierarchy category.
--  @param name the name of the hierarchy category.
--  @return `categoryKey`: the key value for the hierarchy category. If the category cannot be found or the key cannot be retrieved, 
--  `nil` is returned. 
--  @return `pass`: a Boolean value that indicates whether the hierarchy category was found. A value of true indicates success.
function M.hierarchy_category_key_get(name)
   check_args( name,{["fn"]="hierarchy_category_key_get", "string"} )
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return nil, false 
   end
   local categoryKey

   local categories = M.hierarchy_category_get_all()
   for i,v in pairs(categories) do
      if categories[i].name == name then
         categoryKey = categories[i].categoryKey
         return categoryKey, true
      end
   end  
   return nil, false
end
    
--- Retrieve a table that contains all hierarchies categories.
--  @return `categories`: a table with the hierarchy categories. For each hierarchy category, the table includes the 
--  category name and additional information, such as the category key.
function M.hierarchy_category_get_all()
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return
   end
   local categories
   local pass, json_out, code = rest.request("GET", "config/hierarchies/categories", nil, 'Accept:text/plain, */*; q=0.01')
   if pass then
      categories = Json:decode(json_out)
      categories = categories.items
   end
   return categories
end

--- Delete a hierarchy category. 
-- @param name the name of the hierarchy category to delete.
-- @param safe a Boolean value that indicates whether to delete a hierarchy category
-- if the category contains hierarchies. A value of true indicates that the deletion
-- is performed only if there are no hierarchies in the category. A value of false 
-- indicates that the category and any hierarchies that it contains are deleted. 
-- The default value is true (safe delete).
-- @return `pass`: a Boolean value that indicates whether the category was deleted.
-- @return `msg`: a message from the server with additional information about the operation.
function M.hierarchy_category_delete(name,safe)
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return
   end
   if safe == nil then
      safe = true
   end
   local categories = M.hierarchy_category_get_all()
   local cat_key
   for _,category in ipairs(categories) do
      if category.name:lower() == name:lower() then
         cat_key = category.key
         break
      end
   end
   if not cat_key then 
      return true, "Category did not exist" 
   end
   if safe then
      local empty = true
      local hierarchies = M.hierarchy_get()
      for _,h in ipairs(hierarchies) do
         if h.categoryKey == cat_key then
            return false,'Category was not empty'
         end
      end
   end
   local pass, json_out, code = rest.request("DELETE", "config/hierarchies/categories/"..tostring(cat_key))
   return pass, json_out
end

--- Retrieve the line-item key for a specific hierarchy member.
--  @param Hierarchy_Name the hierarchy to search.
--  @param Version the version of the hierarchy.
--  @param Member_name the hierarchy member for which to retrieve the line-item key.
--  @return `li_key`: the line-item key for the specified member. If the key is not found, `nil` is returned.
function M.hierarchy_li_key_get(Hierarchy_Name, Version, Member_name)
   check_args( Hierarchy_Name, Version, Member_name,{["fn"]="hierarchy_li_key_get", "string","string","string"} )
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return
   end
  local li_key
  local pass = rest.request( "GET", "config/hierarchies/details", nil, "Accept:text/plain, */*; q=0.01")
  if not pass then
     sas.print("%1z"..tostring(utils.read_fileref("hout")))
  end  
  local hier_details = Json:decode(utils.read_fileref("out"))
  for i,v in ipairs(hier_details.items) do
     if not v.hierarchy.version then v.hierarchy.version = "" end
     if sas.lowcase(v.hierarchy.version) == sas.lowcase(Version) and sas.lowcase(v.hierarchy.name) == sas.lowcase(Hierarchy_Name) then
        for j,w in pairs(v.lineItems) do
           if w.memberName == Member_name then
              return w.liKey
           end
        end
     end
  end
  return
end

--- Delete a hierarchy.
-- 
--  **Note:** Because the object cannot be recovered, use caution when you delete a hierarchy.
--  @param Name the name of the hierarchy to delete.
--  @param Version the version of the hierarchy to delete.
--  @return `pass`: a Boolean value that indicates whether the hierarchy was deleted. A value of true indicates success.
--  @return `msg`: a message from the server with additional information about the operation.
function M.hierarchy_delete(Name, Version)
   check_args( Name, Version,{["fn"]="hierarchy_delete", "string",'string'} )
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return false, nil
   end
   check_args(Name, Version, {["fn"]="hierarchy_delete", "string","string"})
   local id, pass = M.hierarchy_key_get(Name, Version)
   if not pass or not id then return true,nil end
   local pass = rest.request( "DELETE", "config/hierarchies/"..tostring(id), nil, "Accept:*/*\nAccept-Encoding: gzip,deflate")
   return pass,utils.read_fileref('out')
end

-- *****************************************************
-- *                    #SCENARIO                      *
-- *****************************************************

--- Scenario Functions
-- @section Scenario

--- Add a scenario to a project.
--  @param project the name of the project that you want to add scenarios to.
--  @param scenario the scenario to add to the project. You can specify this either as a string that contains the scenario name or as
--  a table that specifies the scenario name and the scenario version. For example, if `arg` is a table, `arg.name` is the scenario name, and `arg.version`
--  is the scenario version.
--  @return `pass`: a Boolean value that indicates whether the scenario was added. A value of true indicates success.
function M.scenario_add_to_project(project, scenario)
   check_args( project, scenario,{["fn"]="scenario_add_to_project", "string",{"string","table"}} )
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return false 
   end
   local key = M.project_key_get(project)   
   if not key then
       -- sas.print("%1z Could not find the project "..project)
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_scenario_add_to_project_1', 'NOQUOTE'),project)
      return false
   end   
 
   local body = M.scenario_make_add_to_project_string(scenario)
   if not body then
       -- sas.print("%1z Could not find the scenario "..scenario)
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_scenario_add_to_project_2', 'NOQUOTE'),scenario)
      return false
   end
   body = "["..body.."]"
   local header_in ='Accept:application/json, text/javascript, */*; q=0.01 \nAccept-Encoding:gzip, deflate \nContent-Item:application/vnd.sas.risk.scenario.summary+json;charset=utf-8'

   local pass, json_out, code = rest.request( 'POST', 'projects/'..key..'/scenarios', body, header_in, 'application/vnd.sas.collection+json;charset=UTF-8', nil)
   return pass
end

--- Import a scenario into the scenario repository.
--  @param file the full path to the Microsoft Excel file that defines the scenario to be imported. 
--  @param creation_date the creation date for the scenario. This date identifies the last time horizon in the scenario that contains historical data.
--  The value for this parameter is used only if the value of `var_history` is true.
--  @param var_history a Boolean value that specifies whether to create variable history for the scenario variables during the import process. 
--  The default value is false.
--  @return `pass`: a Boolean value that indicates whether the scenario was imported. A value of true indicates success.
--  @return `out_str`: an informative message that is returned from the HTTP request.
function M.scenario_create(file, creation_date, var_history)
   check_args( file, creation_date, var_history,{["fn"]="scenario_create", "string",{"string",'nil'},"boolean"} )
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return false 
   end
   local file_sep = utils.get_file_sep()
   local file_name = sas.scan(file,-1,file_sep)
   local url_str = 'scenarios'
   if var_history then
      url_str = url_str..'?historyDate='..creation_date
   end

   --Create the input header and body
   -- HEADER: 
   local header = 'Accept:*/*\r\nAccept-Encoding:gzip, deflate'

   --BODY:
   local topText =    '------WebKitFormBoundaryi88cnGn8y3Fy4rgp\r\n'
   topText = topText.. 'Content-Disposition: form-data; name="file"; filename="'..file_name..'"\r\n'
   topText = topText.. 'Content-Type: application/vnd.openxmlformats-officedocument.spreadsheetml.sheet\r\n\r\n'

   local bottomText =    '\r\n------WebKitFormBoundaryi88cnGn8y3Fy4rgp\r\n'
   bottomText = bottomText.. 'Content-Disposition: form-data; name="_charset_"\r\n\r\n'
   bottomText = bottomText.. 'UTF-8\r\n'
   bottomText = bottomText.. '------WebKitFormBoundaryi88cnGn8y3Fy4rgp\r\n'
   bottomText = bottomText.. 'Content-Disposition: form-data; name="file-data"\r\n\r\n\r\n'
   bottomText = bottomText.. '------WebKitFormBoundaryi88cnGn8y3Fy4rgp--'

   local xl_logical = sasxx.assign(file,"xl")
   local in_logical = sasxx.new("in")

   local body = topText

   local handle = xl_logical:open('r-o','fc')
   if not handle then
--       sas.print("%1z Could not open "..file..". Check that it exists and is readable")
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_scenario_create_1', 'NOQUOTE'),file)
   end
   for line in handle:lines() do
      body=body..line
   end
   handle:close()
   body = body..bottomText
   handle = in_logical:open('w+o',nil,string.len(body))
   handle:write(body)
   handle:close()

   xl_logical:deassign()
       
   local pass, scen_string, code = rest.request("POST", url_str, {[1]='in'}, header, 'multipart/form-data; boundary=----WebKitFormBoundaryi88cnGn8y3Fy4rgp', nil)
   if not pass then return false,scen_string end

   --The output from this post gets passed in to the next.
   local scen_table = Json:decode(scen_string)
   local scen_body
   if scen_table.items[1][1] then
      scen_table = scen_table.items[1]
   else
      scen_table = scen_table.items[2]
   end
   local scen_body = Json:encode(scen_table)
   utils.write_fileref('in',scen_body)
   sas.submit([[
      data _null_;
         file hin;
         put 'Accept:application/vnd.sas.collection+json;charset=utf-8, */*; q=0.01';
         put 'Accept-Encoding:gzip, deflate';
         put 'Accept-Item:application/vnd.sas.risk.scenario.summary+json;charset=utf-8';
         put 'Content-Item:application/vnd.sas.risk.scenario+json;charset=utf-8';
      run; 
   ]])  
   local pass, out_str, code = rest.request("POST", "scenarios?update=true", {[1]='in'}, {[1]='hin'}, 'application/json', nil)   
   return pass, out_str
end

--- Retrieve detailed information about a scenario. This function is used by the `scenario_add_to_project` function to add a scenario to a project.
--  @param scenario the scenario for which to retrieve information. You can specify this either as a string that contains the scenario name or as
--  a table that specifies the scenario name and the scenario version. For example, if `arg` is a table, `arg.name` is the scenario name, and `arg.version`
--  is the scenario version.
--  @return A string that can be used as input to the `scenario_add_to_project` function.
function M.scenario_make_add_to_project_string(scenario)
   check_args( scenario,{["fn"]="scenario_make_add_to_project_string", {"string","table"}} )
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return nil
   end
   local name, version
   if type(scenario) == 'string' then
      name = scenario
      version = ''
   else
      name = scenario.name
      version = scenario.version
   end
   local header = 'Accept:application/vnd.sas.risk.scenario.summary+json;charset=utf-8, */*; q=0.01\nAccept-Encoding:gzip, deflate \nAccept-Item:application/vnd.sas.risk.scenario+json;charset=utf-8'
   local pass, out_str, code = rest.request('GET', 'scenarios', nil, header, 'application/json;charset=UTF-8') 
   if not pass then return nil end   
   local output_table = {}
   local scenarios = Json:decode(out_str)
   for i,scenario in pairs(scenarios.items) do
      if scenario.label == name and scenario.scenarioVersion == version then
         output_table.id = scenario.id
         output_table.name = scenario.name
         output_table.periodType = scenario.periodType
         output_table.baselineId = scenario.baselineId
         output_table.label = scenario.label
         output_table.filename = scenario.filename
         output_table.readOnly = scenario.readOnly
         output_table.deprecated = scenario.deprecated
         output_table.periods = scenario.periods
         output_table.stresses = scenario.stresses
         output_table.sourceId = scenario.sourceId
         output_table.modifiedBy = scenario.modifiedBy
         output_table.modifiedTimeStamp = scenario.modifiedTimeStamp
         output_table.createdBy = scenario.createdBy
         output_table.creationTimeStamp = scenario.creationTimeStamp
         output_table.version = scenario.version
         output_table.scenarioVersion = scenario.scenarioVersion
         output_table.links = scenario.links
         found = true
         return Json:encode(output_table)
      end   
   end
   return nil
end
   
--- Retrieve the key value that identifies a scenario.
--  @param scenario the name of the scenario. You can specify this either as a string that contains the scenario name or as
--  a table that specifies the scenario name and the scenario version. For example, if `arg` is a table, `arg.name` is the scenario name, and `arg.version`
--  is the scenario version.
--  @param project_id the numeric identifier for the project that contains the scenario that you are interested in. 
--  @return `key`: the key value for the scenario. If the scenario cannot be found or the key cannot be retrieved, `nil` is returned. 
function M.scenario_key_get(scenario, project_id)
   check_args( scenario, project_id,{["fn"]="scenario_key_get", {"string","table"},"number"} )
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return nil
   end
   local name, version
   if type(scenario) == 'string' then
      name = scenario
      version = ''
   else
      name = scenario.name
      version = scenario.version
   end
   local pass, str, code = rest.request('GET', 'projects/'..project_id..'/scenarios')
   local scenarios = Json:decode(str)
   for i,v in pairs(scenarios.items) do
      if v.name == name and v.scenarioVersion == version then
         return v.scenarioKey
      end
   end
   return nil
end

--- Retrieve the line-item key that identifies a scenario in SAS Risk and Finance Workbench. The line-item key represents the scenario in crossings that 
--  identify data points in the system.
--  @param scenario the name of the scenario. You can specify this either as a string that contains the scenario name or as
--  a table that specifies the scenario name and the scenario version. For example, if `arg` is a table, `arg.name` is the scenario name, and `arg.version`
--  is the scenario version.
--  @param project_key the numeric identifier for the project that contains the scenario that you are interested in. 
--  @return The line-item key for the scenario. If the scenario cannot be found or the key cannot be retrieved, `nil` is returned. 
function M.scenario_likey_get( scenario, project_key )
   check_args(scenario, project_key,{["fn"]="scenario_likey_get", {"string","table"},"number"} )
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return nil
   end
   if not M.url_set() then return nil end
   local name, version
   if type(scenario) == 'string' then
      name = scenario
      version = ''
   else
      name = scenario.name
      version = scenario.version
   end
   local pass, json = rest.request("get", "projects/" .. project_key .. "/scenarios/")
   if not pass then
--       sas.print("%1zCan't find liKey for scenario <%s>.", scenario_name)
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_scenario_likey_get_1', 'NOQUOTE'), scenario_name)
--       sas.print("%1zResponse %s.", json)
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_scenario_likey_get_2', 'NOQUOTE'), json)
      return
   end
   local scen = Json:decode(json)
   
   for i ,h in ipairs(scen.items) do
      if h.name == name and h.scenarioVersion == version then
         return h.liKey
      end     
      if version == '' and h.name == name then
         return h.liKey
      end
   end
   return nil
end
---Retrieve a list of all the scenario names in a project.
--  @param projectId the numeric identifier of the project. 
--  @return `scenario_names`: a table of the scenario names in the specified project. If there are no scenarios, `nil` is returned.
function M.scenario_names_get( projectId )
   check_args( projectId,{["fn"]="scenario_names_get", "number"} )
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return nil
   end
-- get the scenario names for this project
    local json, pass
    pass, json = rest.request( "get", "projects/"..projectId .. "/scenarios" )
    if not pass then
--           sas.print("%1zCan't get a list of scenarios from project id %s. ", projectId )
       sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_scenario_names_get_1', 'NOQUOTE'), tostring(projectId) )
       return nil
    end
    local scens = Json:decode(json  )
    
    if not scens.items then return nil end
    local scenario_names = {}
    for i, scn in ipairs(scens.items) do
        scenario_names[#scenario_names +1] = scn.name
    end
    return scenario_names
end

--- Delete a Scenario
-- @param name - The name of the scenario to delete
-- @param version - The version of the scenario. Defaults to ""
-- @return pass - whether or not the delete was succesful
function M.scenario_delete( name, version )

   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return nil
   end
   
   version = version or ""
   -- Get all the scenarios
   local pass, json_out, code = rest.request( 'GET', 'scenarios') 
   local scen_all = Json:decode(json_out)
   for _,scen in pairs(scen_all.items) do
      if not scen.scenarioVersion then scen.scenarioVersion = "" end
      if scen.name == name and tostring(scen.scenarioVersion) == tostring(version) then
         for _,link in pairs(scen.links) do
            if link.rel == 'delete' then
               Debug('deleting '..scen.label or (scen.name.." / "..tostring(scen.scenarioVersion)))
               pass = rest.request( 'DELETE', link.uri)
               return pass
            end
         end
      end
   end
   return false
end

   
-- *****************************************************
-- *                      #MODEL                       *
-- *****************************************************
    
--- Model Functions
--  @section Models

--- Import the models from a specified directory into SAS Risk and Finance Workbench. All models in the directory are imported, 
--  unless a model with the same name already exists in the system.
--  @param Dir_Name the full path to the directory on the server that contains the models to import.
--  @return `pass`: a Boolean value that indicates whether the models were imported. A value of true indicates success.
function M.model_load(Dir_Name)
   check_args( Dir_Name, {["fn"]="model_load", "string"} )
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return false 
   end
   utils.write_fileref('in','baseDir:'..Dir_Name)
   utils.write_fileref('hin','Accept:application/json\nAccept-Encoding:gzip, deflate')
   local pass, out_str, code = rest.request("POST", "models/load?baseDir="..Dir_Name, {[1]='in'}, {[1]='hin'}, 'application/vnd.sas.modinv.rest.model+json')
   return pass
end

--- Retrieve the key value that identifies a model.
--  @param Name the name of the model.
--  @param Version the version of the model.
--  @return `key`: the key value for the model. If the model cannot be found or the key cannot be retrieved, `nil` is returned. 
function M.model_key_get(Name, Version)
   check_args( Name, Version,{["fn"]="model_key_get", "string","string"} )
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return nil
   end
   Version = Version or ""
   local pass, models, code = rest.request("GET", "models/", nil, 'Accept:application/json \nAccept-Encoding:gzip, deflate')
   if not pass then return nil end
   local models_table = Json:decode(models)
   local key = nil
   for i,v in pairs(models_table.items) do
      if v.name == Name and v.version == Version then
         key = v.id
         return key
      end
   end
   return nil
end

-- *****************************************************
-- *                 #MODEL GROUP                      *
-- *****************************************************

--- Model Group Functions
-- @section ModelGroups

--- Create a model group.
--  @param Name the name of the model group to create.
--  @param Version an identifier to distinguish this model group from other model groups with the same name.
--  @param Interval the time period to use for the model group. The following values are supported: `YEAR`, `SEMIYEAR`, `QUARTER`, `MONTH`, `WEEK`.
--  The models that you use with this model group must use the same interval value.
--  @param Description a description for the model group.
--  @param Models a list of models to include in the model group. Separate multiple entries with semicolons.
--  @param Model_Versions a list of version identifiers for the models in the model group. Separate multiple entries with semicolons.
--  @param Logic the user-defined logic that the model group executes.
--  @param Hierarchies a list of hierarchies that contains members the model group generates output for. Separate multiple entries with semicolons.
--  @param Hierarchy_versions a list of version identifiers for the hierarchies. Separate multiple entries with semicolons.
--  @param Classification a table of classification variable names for the output hierarchy.
--  @param Member_names a table of the hierarchy member names that the generated output is written to.
--  @param Outvar_names a table of output variables in the model results that are associated with the hierarchy members.
--  @param Assumption_names a list of business assumption names. Separate multiple entries with semicolons.
--  @param Assumption_values a list of values for the business assumptions. Separate multiple entries with semicolons.
--  @param Assumption_descriptions a list of descriptions for the business assumptions. Separate multiple entries with semicolons.
--  @param update a Boolean value that indicates whether to update the model group if it already exists. The default value is false. If the model group does not already exist,
--  it is created, regardless of this value.
--  @return `pass`: a Boolean value that indicates whether the model group was created. A value of true indicates success.
function M.model_group_create(Name,Version,Interval,Description,Models,Model_Versions,Logic,Hierarchies,Hierarchy_versions,Classification,Member_names,Outvar_names,Assumption_names,Assumption_values,Assumption_descriptions,update)
   check_args( Name,Version,Interval,Description,Models,Model_Versions,Logic,Hierarchies,Hierarchy_versions,Classification,Member_names,Outvar_names,Assumption_names,Assumption_values,Assumption_descriptions,update,{["fn"]="model_group_create", "string",{"string",'nil'},"string","string","string",{"string",'nil'},{"string",'nil'},"string","string","table","table","table","string",{"string",'nil'},{"string",'nil'},{"boolean",'nil'}} )
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return false 
   end
   Version = Version or ""
   -- Check to see if we are updating an existing Model Group or making a new one
   if not update then update = false end
   local request_type = 'POST'
   local mod_grp_id
   local url_append = "modelGroups"
   -- See if the Model Group exists. If update, then change from post to put. If not update and it already exists, exit
   local pass, out_str, code = rest.request("GET", "modelGroups")
   if not pass then return false end
   local mgs = Json:decode(out_str)   
   --Search for the worksheet with the supplied name. 
   for i,v in pairs(mgs.items) do
      if sas.lowcase(v.name) == sas.lowcase(Name) and sas.lowcase(v.version) == sas.lowcase(Version) then
         if update then
            mod_grp_id = v.id
            url_append = url_append .. '/'..tostring(mod_grp_id)
            request_type = "PUT"
            break
         else
            return false
         end
      end
   end   

   -- Get the Model Key for each model. Put them in a semicolon separated list.
   local Model_Keys, Key, i, name
   Model_Keys = ""
   if Models and string.trim(Models) ~= "" then 
      local n_models = sas.countw(Models,";")
      for i=1,n_models do
         local model_name = sas.scan(Models,i,";")
         local model_version = sas.scan(Model_Versions,i,";") or ""
         Key = M.model_key_get(model_name,model_version)
         if not Key then 
            sas.print("%1z Could not find model <"..model_name.."> requested for model group "..Name) 
            return false
         end
         if i>1 then
            Model_Keys = Model_Keys..","
         end
         Model_Keys = Model_Keys..Key
      end
   end

   -- escape any quotes and  new lines for user defined logic and business assumption values   
   Logic = M.escape_json_string_value(Logic)
   Assumption_values = M.escape_json_string_value(Assumption_values)
   Description = M.escape_json_string_value(Description)

   local body = '{"name":"'..Name..'","version":"'..Version..'","description":"'..Description..'","modelKeys":['..Model_Keys..'],"periodType":"'..Interval..'","udlGlueCode":"'..Logic..'","outputVars":['
           
   if Hierarchies and sas.strip(Hierarchies) ~= "" then
      --Loop over all the hierarchies
      local Hierarchy = ""
      local Hierarchy_version = ""
      local n_hierarchies = sas.countw(Hierarchies,';')
      for i=1,n_hierarchies do
         Hierarchy = sas.scan(Hierarchies,i,';')
         Hierarchy_version = sas.strip(sas.scan(Hierarchy_versions,i,';'))
         local hier_key, pass   
         hier_key, pass = M.hierarchy_key_get(Hierarchy,Hierarchy_version)
         if not pass then
--             sas.print("%1z Could not find the hierarchy: "..Hierarchy.." when creating the Model Group "..Name) 
            sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_model_group_create_1', 'NOQUOTE' ),Hierarchy,Name)
            return false
         end      
         local class_list = ""
         local n_class = sas.countw(Classification[i],';')
         for j=1,n_class do
            class_list = class_list..'"'..sas.scan(Classification[i],j,';')..'"'
            if j < n_class then
               class_list = class_list..','
            end
         end
         if i > 1 then
            body = body..','
         end
         body = body..'{"hierarchyId":'..hier_key..',"classificationVarNames":['..class_list..'],"outputVarMappings":['
         if Member_names[i] and sas.strip(Member_names[i]) ~= "" then
            local n_members = sas.countw(Member_names[i],';')
            for j=1,n_members do
               local Member = sas.scan(Member_names[i],j,';')
               local Outvar = sas.scan(Outvar_names[i],j,';')
               body = body..'{"memberName":"'..Member..'","varName":"'..Outvar..'"}'
               if j < n_members then
                  body = body..','
               end               
            end
         end
         body = body..']}'
      end
   end

   body = body..'],"defaultRuntimeBusinessAssumptions":['

   if Assumption_names and sas.strip(Assumption_names) ~= "" then
      local assumption, value, description
      local n_assumptions = sas.countw(Assumption_names,";")
      for i=1,n_assumptions do
         assumption = sas.scan(Assumption_names,i,';')
         value = sas.scan(Assumption_values,i,';')
         description = " "
         if sas.countw(Assumption_descriptions,";") >= i then
            description = sas.scan(Assumption_descriptions,i,';')
         end
         body = body..'{"name":"'..assumption..'","value":"'..value..'","description":"'..description..'"}'
         if i < n_assumptions then
            body = body..','
         end
      end 
   end
   body = body..']}'
   
   local pass, out_str, code = rest.request(request_type, url_append, body, 'Accept:application/json \nAccept-Encoding:gzip, deflate', 'application/json;charset=utf-8')  
   return pass
end

---Retrieve cross-classification variables and a hierarchy map for a model group.
--  @param projectId the numeric identifier for the project in which you are running the model group.
--  @param modelgroup the name of the model group.
--  @param modelGroupVersion the version of the model group.
--  @return `ccvs`: a string that contains a list of the cross-classification variables, separated by spaces. If no cross-classification variables are found, `nil` is returned. 
--  @return `hierarchyMap`: a table that maps hierarchy level names to cross-classification variables. For example, `{hierarchyName1 = {ccvName1, ccvName2, ..} , 
--  hierarchyName2 = {ccvName1, ccvName2, ..} }`.
--  @return `json`: a JSON-formatted string that contains the hierarchy map for serialization.
--  @return `interval`: the time period used by the model group (for example, `Quarter`).
function M.ccvs_hierarchy_get(projectId, modelgroup, modelGroupVersion)
   check_args( projectId, modelgroup, modelGroupVersion,{["fn"]="ccvs_hierarchy_get", "number","string","string"} )
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return
   end
    modelGroupVersion = modelGroupVersion or ""
    
  -- get the modelgroups names for this project
    local pass, json = rest.request("get", "modelGroups/")

      if not pass then 
--           sas.print("%1zCan't get a list of modelgroups. ")
       sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_ccvs_hierarchy_get_1', 'NOQUOTE'))
       print(json)
       return nil
    end

    local mgs = Json:decode(json )  
    local mg
    -- Look for the model group in the list
    -- Debug(table.tostring(mgs))
    for i, m in ipairs(mgs.items) do
       if m.name == modelgroup and m.version == modelGroupVersion then
          mg = m
          break
       end      
    end
    
    if not mg then
--           sas.print("%1zCan't find %s in the list of model groups.", modelgroup)
       sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_ccvs_hierarchy_get_2', 'NOQUOTE'), modelgroup .. " (version: " .. modelGroupVersion .. ")")
       return nil
    end
    Debug(mg)
    
    -- extract out the ccvs and make the hierarchy map
    local ccvs = ""
    local hierarchyMap = {}
    for i, out in ipairs(mg.outputVars) do
         -- get the hierarchy details
        local pass, json = rest.request("get", "config/hierarchies/"..out.hierarchyId)
        if not pass then
--                sas.print("%1zCan't get hierarchy details for id %d. ", out.hierarchyId )
            sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_ccvs_hierarchy_get_3', 'NOQUOTE'), out.hierarchyId )
            print(json)
            return nil
        end
        local hier = Json:decode(json )  
        Debug(hier)
        hierarchyMap[hier.name] = {}
        hierarchyMap[hier.name].classificationVarNames = {}
        -- Set the hierarchy version
        hierarchyMap[hier.name].version = hier.version
        -- loop over the level names and class var names and associate them
        for j, ln in ipairs(hier.levelNames) do
            hierarchyMap[hier.name].classificationVarNames[j] = out.classificationVarNames[j]
            hierarchyMap[hier.name].hierarchyId = out.hierarchyId
            ccvs = ccvs.. " " .. out.classificationVarNames[j]
        end
    end
    -- add entity map and scenario map - two level entity allowed
    -- Get the entity hierarchy name
    -- check for the name of entity hierarchy being provided
    local ent = M.entity_hierarchy_get(sas.symget("sst_current_project_id"))
    local hsymVersion = ent.hierarchy.version
    local hsym = ent.hierarchy.name or 'Entity'
    hierarchyMap[hsym] = {}
    hierarchyMap[hsym].classificationVarNames = {"entity", "entity"}
    hierarchyMap[hsym].version = hsymVersion
    hierarchyMap[hsym].hierarchyId = ent.hierarchy.hierarchyKey
    
    hierarchyMap.Scenario = {}
    hierarchyMap.Scenario.classificationVarNames = {"scenario_name"}
    json = Json:encode(hierarchyMap)     
    return ccvs, hierarchyMap, json, mg.periodType
end
   
--- Retrieve the key value that identifies a model group.
--  @param name the name of the model group.
--  @param version the version for the model group. If the model group has no version, specify `""`. 
--  @return `key`: the key value for the model group. If the model group is not found, `nil` is returned. 
function M.model_group_key_get(name, version)
   check_args( name, version,{["fn"]="model_group_key_get", "string","string"} )
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return
   end
   version = version or ""
   local pass, str, code = rest.request("GET", "modelGroups")
   local model_groups = Json:decode(str)
   local key = nil
   for i,v in pairs(model_groups.items) do
      if sas.lowcase(v.name) == sas.lowcase(name) and sas.lowcase(v.version) == sas.lowcase(version) then
         key = v.key
         return key
      end
   end
   return nil
end
   
--- Delete a model group.
-- 
--  **Note:** Because the object cannot be recovered, use caution when you delete a model group.
--  @param Name the name of the model group to delete.
--  @param Version the version of the model group to delete.
--  @return `pass`: a Boolean value that indicates whether the model group was deleted. A value of true indicates success.
--  @return `msg`: a message from the server with additional information about the operation.
function M.model_group_delete(Name, Version)
   check_args(Name, Version, {["fn"]="model_group_delete", "string","string"})
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return false, nil
   end
   local id = M.model_group_key_get(Name, Version)
   if not id then return true,nil end
   local pass = rest.request( "DELETE", "modelGroups",'['..tostring(id)..']', "Accept:application/json\nAccept-Encoding: gzip,deflate",'application/json;charset=UTF-8')
   return pass,utils.read_fileref('out')
end
   
-- *******************************************************
-- *                 #MODEL GROUP RUN                    *
-- *******************************************************
-- @section Model Group Runs

--- Create a model group run
-- @param project_name_or_key - string or number, the project key or name for the project to create the model group in
-- @param name - string, Name of the model group run
-- @param description - string, The description
-- @param work_group - string, name of the workgroup
-- @param scenario[opt] - string, name of the scenario (only 1). pass nil for no scenario.
-- @param entity - string, name of the entity (only 1)
-- @param include_debug boolean, true is same as checking the include debug information box
-- @param model_group - table, name and version of the model group (only 1). e.g. {name="My mg",version="v1"}
-- @param assumptions - table, business assumption overrides. value is always a string. e.g. {{name ="foo", value="bar"},{name ="foo2", value="3"}}. Pass nil for no overrides
-- @return pass - boolean, whether or not the request was succesful
-- @return body_out - the returned body from the request

function M.model_group_run_create(project_name_or_key,name,description,work_group,scenario,entity,include_debug,model_group,assumptions)
   Trace("model_group_run_create",project_name_or_key,name,description,work_group,scenario,entity,include_debug,model_group,assumptions)

   check_args( project_name_or_key,name,description,work_group,scenario,entity,include_debug,model_group,assumptions, {["fn"]="model_group_run_create",{"string","number"},"string","string","string",{"string","nil"},"string","boolean","table",{"table","nil"}} )
   
   local projKey = ""
   if type(project_name_or_key) == 'string' then
      projKey = M.project_key_get(project_name_or_key)
      if not projKey then    
      --'%1zCould not find the project named '..project_name_or_key
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_project_key_get_1', 'NOQUOTE'),project_name_or_key)
      end
   else
      projKey = project_name_or_key
   end
   local mg_key = M.model_group_key_get(model_group.name, model_group.version)
   if not mg_key then
      --'%1zCould not find the model group key for: '..table.tostring(model_group)
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_ccvs_hierarchy_get_2', 'NOQUOTE'),table.tostring(model_group))
      return false
   end
   local entityLIKey = M.entity_likey_get(entity, projKey)
   if not entityLIKey then
      --'%1zCould not find the key for entity named '..entity))
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_model_group_agg_load_setup_3', 'NOQUOTE'),entity)
      return false
   end
   local scenarioKey
   if not scenario then
      scenarioKey = "null"
   else
      scenarioKey = M.scenario_key_get(scenario, projKey)
      if not scenarioKey then
         --'%1zCould not find the key for scenario named '..scenario
         sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_scenario_likey_get_1', 'NOQUOTE'),scenario)
         return false
      end
   end
   local debug_num = "0"
   if include_debug then debug_num = "1" end
   
   local assumptions_json = ""
   if assumptions then
      assumptions_json = Json:encode(assumptions)
   else
      assumptions_json = '[]'
   end
   
   local runTypeStr = "Standard"
   local runMode = "DoNotRun"
  
   local body = string.resolve('{"runInputs":[{"modelGroupKey":@mg_key@,"name":"@name@","description":"@description@","runType":"@runTypeStr@","debug":@debug_num@,"workGroupName":"@work_group@","overriddenAssumptions":@assumptions_json@,"runTypeStr":"@runTypeStr@","entityLIKey":"@entityLIKey@","scenarioKey":@scenarioKey@}],"runMode":"@runMode@"}')
   
   local header = 'Accept: application/json\nAccept-Encoding: gzip, deflate'
  
   projKey = tostring(projKey)
   local url = string.resolve('modelGroups/runs?projectKey=@projKey@&workGroupName=@work_group@&howToRun=@runMode@')
  
   local pass,body_out = rest.request('POST',url,body,header,'application/json;charset=UTF-8')
   return pass,body_out
end

--- Create a model group run for a specified type of detail data. SAS Risk and Finance Workbench uses the type of detail data to determine the following:
-- 
--  * the name of the model group run that is created (format: Load\_Data\__Detail-Kind_\_Run)
--  * the name of the model group that is associated with the model group run (format: Load\_Data\__Detail-Kind_)
--
--  For example, for a detail type of `Loans`, the function generates a model group run with the name Load\_Data\_Loans\_Run and that uses a model group 
--  with the name Load_Data_Loans.
-- @param project_key the project key for the current project.
-- @param detail_kind the type of detail data for which to generate model group runs.
-- @param entities an integer-indexed table with the names of the entities for which to generate model group runs.
-- @param scenarios[opt] an integer-indexed table with the names of the scenarios for which to generate model group runs.
-- @return a Boolean value that indicates whether the model group run was created. A value of true indicates success. 
function M.model_group_agg_load_setup(project_key, detail_kind, entities, scenarios)
   check_args( project_key, detail_kind, entities, scenarios,{["fn"]="model_group_agg_load_setup", "number","string","table",{"table","nil"}} )
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return false 
   end

   -- May not have scenarios
   if not scenarios then
      scenarios = {[1] = ''}
   end

   -- The name we are looking for is Load_Details_<detail_kind>
   local modelGroupName = "Load_Details_"..detail_kind

   -- First get the model group information
   local pass, json_out = rest.request( "get", "modelGroups")
   local model_groups = Json:decode(json_out)
           
   local found = false
   local index = 0
   local model_group_key = 0
   -- Look for the model group with this name
   for i,mg in pairs(model_groups["items"]) do
      if mg["name"] == modelGroupName then
         found = true
         index = i
         model_group_key = mg["id"]
         break
      end
   end
       
   if not found then
--       sas.print("%1zUnable to find the model group "..modelGroupName)
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_model_group_agg_load_setup_1', 'NOQUOTE'),modelGroupName)
      return false
   end
           
   -- Got through all the scenario / entity combos. For each, check if the MGR already exists. If it does, replace it. Otherwise
   -- create a new one.
   local header = 'Accept:application/json\nAccept-Encoding:gzip, deflate'
   local ct = 'application/json;charset=UTF-8'
   local query_modelGroupName = sas.tranwrd(modelGroupName,' ','%20')
   local request = 'modelGroups/runs?projectKey='..project_key   
   local ind = 0
   for i,ent in pairs(entities) do
      for k,scen in pairs(scenarios) do   
         local request_type
         local scen_arg = scen
         if scen_arg == '' then scen_arg = nil end
         local mgr = M.model_group_run_get_properties(project_key,modelGroupName,ent,scen_arg)
         if mgr then 
            local scen_part
            if scen ~= '' then scen_part = " and scenario "..scen else scen_part = '' end
            --  sas.print("%2zThe model group run %s for entity %s%s already exists",modelGroupName,ent,scen_part)
            sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_model_group_agg_load_setup_2', 'NOQUOTE'),modelGroupName,ent,scen_part)
            return true
         end
                    
         local entity_likey = M.entity_likey_get(ent, project_key)
         if not entity_likey then
            --         sas.print("%1zThe entity %s doesn't exist in this project.", ent)
            sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_model_group_agg_load_setup_3', 'NOQUOTE'), ent)
            return false
         end

         local scenario_likey
         if scen and scen ~= '' then
            scenario_likey = M.scenario_key_get(scen, project_key)
            if not scenario_likey then
               -- sas.print("%1zThe scenario %s doesn't exist in this project.", scen)
               sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_model_group_agg_load_setup_4', 'NOQUOTE'), scen)
              return false
            end
            else
              scenario_likey = "null"
            end
            local body = '{"runSerially":false,"runInputs":['..'{"modelGroupKey":'..model_group_key..
',"name":"'..modelGroupName..'","description":"Placeholder for '..modelGroupName..
' woksheet pushes","runType":"Standard","debug":0,"overriddenAssumptions":[],"runTypeStr":"Standard","entityLIKey":"'..entity_likey..
'","scenarioKey":'..scenario_likey..'}]}'
            Debug(request)
            Debug(body)
            Debug(header)
            local pass, out_str, code = rest.request( 'POST', request, body, header, ct )
            if not pass then
            -- sas.print("%1zAttempt to create the model group run %s failed: %s",modelGroupName,out_str)
            sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_model_group_agg_load_setup_5', 'NOQUOTE'),modelGroupName,out_str)
            return false
         end
      end
   end
   return true
end

--- Get the output directory and other details from a model group run
--  @param project_key the numeric identifier for the project that contains the model group run that you are interested in. 
--  @param name  The name of the model group run  
--  @param entity  The name of the enitity
--  @param scenario[OPTIONAL] The name of the scenario. Pass nil for no scenario.
--  @param model_group [OPTIONAL] Table with the name and version of the model group. e.g. {name="myMGname",version="v3.1415"}. Pass nil for no model group.
--  @return properties  A table with the model group run properties:
--
--  * outputFiles (JSON array)
--  * runningTimeInSeconds (number)
--  * entityHierarchyKey (number)
--  * resultsHaveBeenLoaded (Boolean)
--  * started (timestamp string)
--  * id (number)
--  * overriddenAssumptions (JSON array)
--  * outputDir (string)
--  * scenarioKey (string)
--  * entityLIKey (string)
--  * runByUserName (string)
--  * workgroupName (string)
--  * objectType (string)
--  * projectKey (number)
--  * ended (timestamp string)
--  * asOfDate (string)
--  * modelGroupKey (number)
--  * partialDataLoad (Boolean)
--  * countOfDataValuesLoaded (number)
--  * updated (timestamp string)
--  * runKey (number)
--  * debug (number)
--  * name (string)
--  * runByUserId (string)
--  * status (string)
--  * shocks (JSON array)
--  * created (timestamp string)
--  * runType (string)
--  * errors (JSON array)
--  * description (string)
--
function M.model_group_run_get_properties(project_key,name,entity,scenario,model_group)
   check_args( project_key,name,entity,scenario,{["fn"]="model_group_run_get_properties","number","string","string",{"string","nil"},{"table","nil"}})
   if not M.url_set(nil) then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return
   end
   
   -- Get the liKeys for the scenario and entity
   local scenario_likey
   if scenario then
      scenario_likey = M.scenario_key_get(scenario, project_key)
      if not scenario_likey then
--          sas.print("%1zUnable to find the scenario %s.",scenario)
         sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_model_group_run_get_properties_1', 'NOQUOTE'),scenario)
         return
      end
   end

   local entity_likey = M.entity_likey_get(entity, project_key)
   if not entity_likey then
--       sas.print("%1zUnable to find the entity %s.",entity)
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_model_group_run_get_properties_2', 'NOQUOTE'),entity)
      return
   end

   -- Get the model group key
   local mg_key
   if model_group then
      mg_key = M.model_group_key_get(model_group.name, model_group.version or "")
      if not mg_key then
         --'%1zCould not find the model group key for: '..table.tostring(model_group)
         sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_ccvs_hierarchy_get_2', 'NOQUOTE'),table.tostring(model_group))
         return
      end
   end
   
   -- Get the model group run info and find the correct one to return      
   local pass, json_out, code = rest.request( "get", 'modelGroups/runs?projectKey='..project_key)
   if not pass then
--       sas.print("%1z Unable to get model group run information for this project: "..rest.http_codes[code])
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_model_group_run_get_properties_3', 'NOQUOTE'),rest.http_codes[code])
      return
    end
   local MGRs = Json:decode(json_out)
   Debug(MGRs)
   Debug(name)
   Debug(entity)
   Debug(entity_likey)
   Debug(scenario)
   Debug(scenario_likey)
         
   for i,mgr in pairs(MGRs.items) do 
      if mgr.name == name and mgr.entityLIKey == entity_likey and (not scenario or mgr.scenarioKey == scenario_likey) and (not mg_key or mgr.modelGroupKey == mg_key) then
         return mgr
      end
   end
   return
end

--- Save data in a worksheet to a sas dataset
--  @param project_name_or_key - string or number, Name of the project or the project key
--  @param model_group_run - string, The name of the model_group_run to execute
--  @param scenario[OPTIONAL] - string, The name of the scenario to execute the MGR for. Pass nil for no scenario
--  @param entity - string, The name of the entity to get data for
--  @param model_group[OPTIONAL] - Table with the name and version of the model group. e.g. {name="myMGname",version="v3.1415"}. Pass nil for no model group.
--  @param max_wait_time - number, maximum amount of time (in seconds) to wait for the model group run to finish (run synchronously). To not wait (run asynchronously), pass nilas a tool for testing. In practice it should be left nil.
--  @return pass - boolean, whether or not the function call succeeded. 
--  @return runKey - number, the runkey for the run instance
--  @return updatedTimestamp - The updated time stamp
--  @return msg - string, the returned body from the request to run the MGR
function M.model_group_run_execute(project_name_or_key, model_group_run, scenario, entity, model_group, max_wait_time )
Trace("model_group_run_execute", project_name_or_key, model_group_run, scenario, entity, model_group, max_wait_time )
check_args( project_name_or_key, model_group_run, scenario, entity, model_group, max_wait_time ,{["fn"]="model_group_run_execute", {"string","number"}, "string", {"string","nil"}, "string",{"table","nil"},{"number","nil"}} )
  
   -- Get the project key
   local projKey
   if type(project_name_or_key) == 'string' then
      projKey = M.project_key_get(project_name_or_key)
      if not projKey then
      --'%1zCould not find the project named '..project_name_or_key
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_project_key_get_1', 'NOQUOTE'),project_name_or_key)
      end
   else
      projKey = project_name_or_key
   end
   
   local mgr = M.model_group_run_get_properties(projKey,model_group_run,entity,scenario,model_group)
   if not mgr then
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_model_group_run_2', 'NOQUOTE'),model_group_run)
      return false
   end
   Debug("Model Group Run is:")
   Debug(table.tostring(mgr))
   
   -- Now run it and load values
   pass, json = rest.request( "post", "modelGroups/runs/"..mgr.runKey)
   if not pass then
--     sas.print("%1zRun of  model group run %s wasn't successful. ", model_group_run )
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_model_group_run_3', 'NOQUOTE'), model_group_run )
      return pass, nil, nil, json
   end
   
  -- Wait for the model group to finish
  if not max_wait_time then
     return true, nil, nil, json
  else
     local updatedTimestamp = 0
     local notready = true
     local status, runKey
     local waitseconds = 0
      while notready and waitseconds < max_wait_time do
         sas.sleep(1,1)
         local mgr_post = M.model_group_run_get_properties(projKey,model_group_run,entity,scenario,model_group)
         Debug("Polling:" ,waitseconds)
         status = mgr_post.status
         if status == "Completed" or status == "Error" then
            notready = false
            runKey = mgr_post.runKey
         end
         waitseconds = waitseconds + 1
         updatedTimestamp = sas.put(mgr_post.updated,"D15.")
      end
      Debug("Status = ", status)

      if status == "Error" then
      --     sas.print("%1zExecution of the  model group run %s was unsuccessful. ", model_group_run )
         sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_model_group_run_4', 'NOQUOTE'), model_group_run )
      end

      if waitseconds >  max_wait_time then
      --     sas.print("%1zExecution of the  model group run %s took longer than %d seconds. ", model_group_run, max_wait_time )
         sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_model_group_run_5', 'NOQUOTE'), model_group_run, max_wait_time )
      end

      return true, runKey, updatedTimestamp, json
   end
end


--- Run a model group.
--  @param model_group_run the name of the model group run to execute.
--  @param project_name_or_key the project in which the model group run executes. You can specify the project name as a string, or you can provide 
--  the project key as a number (for example, by using the `project_key_get` function). 
--  @param max_wait_time[opt] the number of seconds to wait for model group run to complete before generating a message. The default is 10 seconds.
--  @return a Boolean value that indicates whether the model group was run. A value of true indicates success. 
--  @return `updatedTimestamp`: the timestamp that identifies when the model group was run.
--  @return `runKey`: the numeric key value for the model group run.
function M.model_group_run( model_group_run, project_name_or_key, max_wait_time )
   Trace("model_group_run", model_group_run, project_name_or_key, max_wait_time )
   check_args( model_group_run, project_name_or_key, max_wait_time,{["fn"]="model_group_run", "string",{"string",'number'},{"number","nil"}} )
   if not M.url_set(nil) then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return
   end
      
  if project_name_or_key == nil then 
--     sas.print("%1zMissing project name." )
    sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_model_group_run_1', 'NOQUOTE' ))
    return false
  end
  
  max_wait_time = max_wait_time or 10

  -- Get the project key if necessary
  local project_key = project_name_or_key
  Debug(project_key)
  Debug(type(project_name_or_key))
  if type(project_name_or_key) == 'string' then
    project_key = M.project_key_get( project_name_or_key)
    if project_key == nil then
      return false
    end
  end

  Debug(project_key)
   
  local json, pass
  pass, json = rest.request( "get", "modelGroups/runs?projectKey=" .. project_key )
  Debug(json)
  local jt = Json:decode(json  )
  Debug(jt.items)

  -- find the most recent run of the named model group run

  local recent_dt = 0
  local mg, mg_index
  for i, rn in ipairs( jt.items) do
    if rn.name == model_group_run and recent_dt  < rn.updated then
      recent_dt = rn.updated
      mg = rn
      mg_index = i
      end      
   end

  if recent_dt == 0 then
--     sas.print("%1zCan't find model group run %s.", model_group_run )
    sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_model_group_run_2', 'NOQUOTE'), model_group_run )
      return false
   end
      
  Debug(mg)
  -- Now run it and load values
  pass, json = rest.request( "post", "modelGroups/runs/"..mg.runKey)

   if not pass then
--     sas.print("%1zRun of  model group run %s wasn't successful. ", model_group_run )
    sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_model_group_run_3', 'NOQUOTE'), model_group_run )
    Debug(json)
   end
   
  -- Wait for the model group to finish
  local notready = true
  local status, runKey
  local waitseconds = 0
  local updatedTimestamp = 0
  while notready and waitseconds < max_wait_time do
    sas.sleep(1,1)
    pass, json = rest.request( "get", "modelGroups/runs?projectKey=" .. project_key )

    local pollit = Json:decode(json  )
    Debug("Model group index:", mg_index )
    Debug("Polling:" ,waitseconds)
    Debug(pollit.items[mg_index])
    status = pollit.items[mg_index].status
    if status == "Completed" or status == "Error" then
      notready = false
      runKey = pollit.items[mg_index].runKey
      updatedTimestamp = pollit.items[mg_index].outputFiles[1].updatedTimestamp;
    end
    waitseconds = waitseconds + 1
  end

  Debug("Status = ", status)

  if status == "Error" then
--     sas.print("%1zExecution of the  model group run %s was unsuccessful. ", model_group_run )
    sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_model_group_run_4', 'NOQUOTE'), model_group_run )
    Debug(json)
   end

  if waitseconds >  max_wait_time then
--     sas.print("%1zExecution of the  model group run %s took longer than %d seconds. ", model_group_run, max_wait_time )
    sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_model_group_run_5', 'NOQUOTE'), model_group_run, max_wait_time )
    Debug(json)
   end

  updatedTimestamp = sas.put(updatedTimestamp,"D15.")
  
  return true, updatedTimestamp, runKey

end

--- Load the results for a model group run.
--  @param model_group_runKey the numeric key value for the model group run.
--  @param updatedTimestamp the value of the updatedTimestamp property for the model group run.
--  
--  You can determine these values with the `model_group_run_get_properties` function.
--  @return a Boolean value that indicates whether the results were loaded. A value of true indicates success. 
function M.model_group_data_load( model_group_runKey, updatedTimestamp )
   Trace("model_group_data_load", model_group_runKey, updatedTimestamp )
   check_args(  model_group_runKey, updatedTimestamp,{["fn"]="model_group_data_load", "number",{"number","string"}} )
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return false 
   end
  if not M.url_set() then  return false end
      
  local json, pass
  pass, json = rest.request( "post", "modelGroups/runs/"..model_group_runKey .. "/projectData?resultsFileUpdatedTimestamp=" .. updatedTimestamp)
  if not pass then
--     sas.print("%1zLoad of data from  model group run %d was unsuccessful. ", model_group_runKey )
    sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_model_group_data_load_1', 'NOQUOTE'), model_group_runKey )
    Debug(json)
    return false
  end
  
  Debug(json)
  return true

end
-- *****************************************************
-- *                     #SCRIPT                       *
-- ****************************************************

--- Script Functions
-- @section Script
  
--- Create a new script.
-- @param Name the name of the script to create.
-- @param Description the description to use for the script.
-- @param Version an identifier to distinguish this script from other scripts with the same name. The default value is "".
-- @param Code the SAS code that you want to use for the script.
-- @param Param_names a list of parameter names that are used by the code. Separate multiple entries with semicolons.
-- @param Param_values a list of values for the parameters. Separate multiple entries with semicolons.
-- @param Param_descriptions a list of descriptions for the parameters. Separate multiple entries with semicolons.
-- @param Param_type a list of types for the parameters. Separate multiple entries with semicolons.
-- @param Param_definition a list of definitions for the parameters. Separate multiple entries with semicolons.
-- @param update[opt] a Boolean value that indicates whether to update the script if it already exists. The default value is false. 
-- @param file - the full path of the excel file
-- If the script does not already exist, it is created, regardless of this value.
-- @return success a Boolean value that indicates whether the script was created. A value of true indicates success.  
function M.script_create( Name,Description,Version,Code,Param_names,Param_values,Param_descriptions,Param_type,Param_definition,update,s_file)
   Trace("script_create", Name,Description,Version,Code,Param_names,Param_values,Param_descriptions,Param_type,Param_definition,update,s_file )
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return false
   end
   
   -- Scan the number of parameter names --
   local num_param_string=""
   local f_num_param_string="FreeForm"  
   local np = sas.countw(Param_names,";")   
   for i=1,(np-1) do
      num_param_string=num_param_string..";" 
      f_num_param_string=f_num_param_string..";FreeForm"
   end
    
   -- Make sure there are values for these parameters, if not make blank entries--
   if Param_type == "" then
      Param_type = f_num_param_string
   end
      
   if Param_definition == "" then
      Param_definition = num_param_string
   end
       
   -- Check to see if we are updating an existing Model Group or making a new one
   update = update or false
   Description = Description or ""
   Version = Version or ""
   local request_type = 'POST'
   local url_append = "scripts"
   -- See if the Script already exists. If it does and update is true, set request to put and adjust the url. If it is there and update is false, abort.
   local pass, out_str, code = rest.request("GET", "scripts")
   if not pass then return false end
   local scripts = Json:decode(out_str)
   for i,v in ipairs(scripts.items) do
     if sas.lowcase(v.name) == sas.lowcase(Name) and sas.lowcase(v.version) == sas.lowcase(Version) then
        if update then
           url_append = url_append.."/"..tostring(v.id)
           request_type = "PUT"
           break
        else
           -- WARNING: Unable to create the script %%s - %%s: it already exists and the update flag is set to false
           sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_script_create_2', 'NOQUOTE'),Name,Version)
           return false
        end
     end
   end

   -- escape any quotes and  new lines for code and business parameter values   
   Code = M.escape_json_string_value(Code)
   Param_values = M.escape_json_string_value(Param_values)
   Param_descriptions = M.escape_json_string_value(Param_descriptions)
   Param_type = M.escape_json_string_value(Param_type)
   Param_definition = M.escape_json_string_value(Param_definition)
   
   local body = '{"name":"'..Name..'","description":"'..Description..'","version":"'..Version..'","code":"'..Code..'","defaultParameters":['

   if Param_names and sas.strip(Param_names) ~= "" then
      local names_tbl = {}
      for w in (Param_names .. ";"):gmatch("([^;]*);") do 
          table.insert(names_tbl, w) 
      end
      local values_tbl = {}
      for w in (Param_values .. ";"):gmatch("([^;]*);") do 
          table.insert(values_tbl, w) 
      end   
      local descriptions_tbl = {}
      for w in (Param_descriptions .. ";"):gmatch("([^;]*);") do 
          table.insert(descriptions_tbl, w) 
      end   
      local type_tbl = {}
      for w in (Param_type .. ";"):gmatch("([^;]*);") do 
          table.insert(type_tbl, w) 
      end
      local definition_tbl = {}
      local param_code_file, code_string, ind
      for w in (Param_definition .. ";"):gmatch("([^;]*);") do 
          if string.find(w,".sas") then 
             ind = sas.findc(s_file,"/\\",'b')
             param_code_file = w
             param_code_file = sas.substr(s_file,1,ind-1)..'/'..param_code_file
             code_string = utils.read_file(param_code_file)   
             code_string = M.escape_json_string_value(code_string)
             table.insert(definition_tbl, code_string)
          else
             table.insert(definition_tbl, w)        
          end
          
      end  

      if #names_tbl ~= #values_tbl or #names_tbl ~= #descriptions_tbl then
         sas.print('%1z Number of parameter names, values, and descriptions must be the same')
         return
      end
      
      -- If the type, definition, and SAS code is defined in the excel sheet, check --
      if #names_tbl ~= #values_tbl or #names_tbl ~= #descriptions_tbl or #names_tbl ~= #type_tbl or #names_tbl ~= #definition_tbl then
         sas.print('%1z Number of parameter names,'.. #names_tbl ..
                   ' values,'.. #values_tbl..
                   ' descriptions,'.. #descriptions_tbl..
                   ' type,'.. #type_tbl..
                   ' definition,'.. #definition_tbl)
         return
      end
           
      for i=1,#names_tbl do
         body = body..'{"name":"'..names_tbl[i]..'","value":"'..values_tbl[i]..'","description":"'..descriptions_tbl[i]..'","typeCd":"'..type_tbl[i]..'","definition":"'..definition_tbl[i]..'"}'                  
         if i < #names_tbl then
            body = body..','
         end
      end
   end
   body = body..']}'
   local pass, out_str, code = rest.request(request_type, url_append, body, 'Accept:application/json \nAccept-Encoding:gzip, deflate', 'application/json;charset=utf-8')  
  if not pass then
--    sas.print("%1z Unable to create the script %s: %s",Name,out_str)
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_script_create_1', 'NOQUOTE'),Name,out_str)
      Debug(body)
    return false
  end

  return true
end

--- Retrieve a list of all scripts on the SAS Risk and Finance Workbench server.
--  @return `scripts`: a table that contains the scripts. If no scripts are found or an error occurs, `nil` is returned.
--  @return `pass`: a Boolean value that indicates whether the list of scripts was retrieved. A value of true indicates success.
--  @return `msg`: a message that contains information from the server about the operation.
function M.script_get_all()
   Trace("script_get_all")
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return
   end
   local pass, json = rest.request( "GET", "scripts")
   local scripts_table
   if pass then scripts_table = Json:decode(json).items end
   return scripts_table,pass,utils.read_fileref('out')
end

--- Retrieve the key value that identifies a script.
--  @param Name the name of the script.
--  @param Version the version identifier that distinguishes this script from other scripts with the same name. The default value is "". 
--  @return `key`: the key value for the script. If the script cannot be found or the key cannot be retrieved, `nil` is returned. 
--  @return `pass`: a Boolean value that indicates whether the key value was retrieved. A value of true indicates success.
--  @return `msg`: a message that contains information from the server about the operation.
function M.script_key_get( Name, Version )
   Trace("script_key_get", Name, Version )
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return  
   end
   check_args(Name, Version, {["fn"]="script_key_get", "string","string"})
   
   local all_scripts,pass,msg = M.script_get_all()
   local msg = utils.read_fileref('out')
   if not pass then return nil,false,msg end
   local key
   for _,script in ipairs(all_scripts) do
      if script.name == Name and script.version == Version then 
         return script.key,true,msg
      end
   end
   return nil,false,msg
end

--- Delete a script.
-- 
--  **Note:** Because the object cannot be recovered, use caution when you delete a script.
--  @param Name the name of the script to delete.
--  @param Version the version of the script to delete.
--  @return `pass`: a Boolean value that indicates whether the script was deleted. A value of true indicates success.
--  @return `msg`: a message from the server with additional information about the operation.
function M.script_delete( Name, Version )
   Trace("script_delete", Name, Version )
   check_args(Name, Version, {["fn"]="script_delete", "string","string"})
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return  
   end
   
   local id,pass,msg = M.script_key_get(Name, Version)
   if not pass then return false,msg end
   if not id then return true,nil end
   local pass = rest.request( "DELETE", "scripts",'['..tostring(id)..']', "Accept:application/json\nAccept-Encoding: gzip,deflate",'application/json;charset=UTF-8')
   return pass,utils.read_fileref('out')
end

-- *****************************************************
-- *                     #ENTITY                       *
-- *****************************************************

--- Entity Functions
-- @section Entity

--- Retrieve the line-item key that is associated with an entity from a specified project.
--  @param entity the name of the entity.
--  @param project_name_or_key the project that you are interested in. You can specify the project name as a string, or you can provide 
--  the project key as a number (for example, by using the `project_key_get` function). 
--  @return `li_key`: the line-item key for the entity. If the key is not found, `nil` is returned. 
function M.entity_likey_get( entity, project_name_or_key )
   Trace("entity_likey_get", entity, project_name_or_key )
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return  
   end

   -- Get the project key if necessary
   local project_key = project_name_or_key
   if type(project_name_or_key) == 'string' then
      project_key = M.project_key_get(project_name_or_key)
      if not project_key then
         return false
      end
   end

   local pass, json_out = rest.request( "get", '/projects/'..project_key..'/entity')
   if not pass then 
--       sas.print("%1zCould not get the enitity details for the project")
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_entity_likey_get_1', 'NOQUOTE'))
      Debug(json_out)
      return nil
   end
   local entities = Json:decode(json_out)
   if not entities.lineItems then
     sas.print("%1zNo sheets are included in the project")
     return
   end
   for i,v in ipairs(entities.lineItems) do
      if v.memberName == entity then
         return v.liKey
      end
   end
   return nil
end


--- Retrieve the hierarchy for an entity that is associated with a specified project.
--  @param project_id the numeric identifier for the project. 
--  @return `hierarchy`: a table that contains the entity hierarchy. The table includes the following elements: 
-- 
--  * hierarchyKey (numeric)
--  * lineItems (table). This element includes the line-item keys and level names for each member of the hierarchy.
--  * hierarchy (table). This element includes the hierarchy name, version, and other details.
function M.entity_hierarchy_get( project_id )
   Trace("entity_hierarchy_get", project_id )
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return  
   end
   
   local pass, json_out = rest.request( "get", '/projects/'..project_id..'/entity')
   if not pass then
--       sas.print("%1z Unable to get the project details")
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_entity_hierarchy_get_1', 'NOQUOTE'))
      return nil
   end
   local hierarchy = Json:decode(json_out)
   return hierarchy
end

-- *****************************************************
-- *                     #REPORT                       *
-- *****************************************************

--- Report Functions
-- @section Reports

--- Create a report category. If no report categories exist, you cannot create a report. Create at least one 
--  category before you attempt to create a report.
--  @param name the name of the report category to create.
--  @param project the name of the project in which you want to create the report category.
--  @return `code`: the HTTP return code for the request.
function M.report_category_create( name, project )
   Trace("report_category_create", name, project )
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return 
   end
   
   -- First get the project ID
   local proj_id = M.project_key_get(project)
   if not proj_id then sas.print("%1z Could not find project "..project) return nil end
   
   local body = '{"projectKey":'..proj_id..',"key":0,"name":"'..name..'","readOnly":false,"reportsCount":0}'
   utils.write_fileref('in','{"projectKey":'..proj_id..',"key":0,"name":"'..name..'","readOnly":false,"reportsCount":0}')
   utils.write_fileref('hin','Accept:application/json \r\nAccept-Encoding:gzip, deflate')
   local pass, out_str, code = rest.request("POST", "config/projects/"..proj_id.."/reportConfigs/categories", {[1]='in'}, {[1]='hin'}, 'application/json;charset=UTF-8')
   return code

  end


--- Create a report in a project.
--  @param name the name of the report to create.
--  @param template the name of the report template that you want to use as the basis for the new report. The template file 
--  must be on the server before you attempt to create a report from it.
--  @param project the name of the project in which you want to create the report.
--  @param category the name of the report category to use for the new report.
--  @return `code`: the HTTP return code for the request.
function M.report_load_into_project( name, template, project, category )
   Trace("report_load_into_project", name, template, project, category )
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return 
   end
   -- First get the project ID
   local proj_id = M.project_key_get(project)
   if not proj_id then sas.print("%1z Could not find project "..project) return nil end
   
   --Get the category key
   local key = M.report_category_key_get(category,project)
   if not key then sas.print("%1z Could not find category "..category) return nil end
   
   utils.write_fileref('in','{"key":0,"projectKey":'..proj_id..',"reportConfigCategoryKey":"'..key..'","reportConfigCategory":"'..category..'","reportMapKey":0,"reportName":"'..name..'","template":"'..template..'","history":[],"generatorName":null,"dataName":null,"dependentTemplate":null,"dependentConfigKey":0,"reportStatus":"","messages":[],"actionControls":{"runnable":true},"taxonomyPackageOwners":[],"canBeEditedIfProjectIsLocked":true}')
   utils.write_fileref('hin','Accept:application/json \nAccept-Encoding:gzip, deflate')
   local pass, out_str, code = rest.request("POST", "config/projects/"..proj_id.."/reportConfigs", {[1]='in'}, {[1]='hin'}, 'application/json;charset=UTF-8')

   return code
   
  end

--- Retrieve the key value that identifies a report category.
--  @param name the name of the report category.
--  @param project the name of the project that contains the category.
--  @return `key`: the numeric identifier for the category. If the category is not found, `nil` is returned.
function M.report_category_key_get( name,project )
   Trace("report_category_key_get", name,project )
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return 
   end
   -- First get the project ID
   local proj_id = M.project_key_get(project)
   if not proj_id then sas.print("%1z Could not find project "..project) return nil end
   
   local pass, out_str, code = rest.request("GET", "config/projects/"..proj_id.."/reportConfigs/categories", nil, nil, 'application/json;charset=UTF-8','out')
   local cats = Json:decode(utils.read_fileref('out'))
   local key = nil
   for i,v in pairs(cats.items) do
      if v.name == name then key = v.key end
   end
   return key
end

---Run one or more reports in a project.
-- @param projectName the name of the project that contains the reports to run.
-- @param reportName the report to run. For a single report, you can specify the report name
-- in quotation marks (for example, `"Sample Report"`). For multiple reports, specify a table that lists the 
-- report names (for example, `{"06_source Report" , "report2", "report3" }`).   
-- @param runComment a string that provides any comments you want to associate with the report run.
-- @param filterValues a table that lists any hierarchy members that are required as filters for the report.
-- For example: `{"Bank 1","Q12015","Baseline_US"}`. For reports that are based on SAS programs or that are included 
-- with filing packages, filters might not be applicable.
-- @param max_wait the number of seconds to wait for the report to complete running before the function exits.
-- The default value is 600 seconds. To exit immediately, set this value to 0.
-- @param wait_interval the number of seconds to wait before checking the current status of the report processing.
-- The default value is 5 seconds.
-- @usage 
-- rfw.report_run("Test project","06_source Report","My batch run comments")
-- rfw.report_run("Test project",{"06_source Report", "report2"} ,"My batch run comments",{"Bank 1","Q12015","Baseline_US"})
function M.report_run( projectName,reportName,runComment,filterValues,max_wait,wait_interval )
   Trace("report_run", projectName,reportName,runComment,filterValues,max_wait,wait_interval )
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return 
   end
   local reportIDs ={}
   local filterArray={}
   max_wait = max_wait or 600
   wait_interval = wait_interval or 5
   runComment=runComment or 'Batch Run'

   if type(reportName) ~= "table" then 
      reportName={reportName}
   end
   
   --Get the project Key for constructing the url
   local projKey = M.project_key_get(projectName)
   if not projKey then 
      sas.print('%3z Project '..tostring(projectName)..' was not found')
      return 
   end

   -- If there are anyfilterValues, take the ones that match filters for this project
   if filterValues ~= nil and filterValues ~= '' then
      local filterValuesStr=Json:encode(filterValues)
      local pass, out_str, code = rest.request( 'GET', "/projects/"..projKey.."/sheets/filterValues", nil, "Accept:*/*; q=0.01",'application/json;charset=UTF-8')
      local projectFilters = Json:decode(out_str)
      for i,v in pairs(projectFilters.items) do 
         for i,v in pairs(v.lineItems) do 
            if string.match(filterValuesStr, v.memberName) then 
               --Debug(i,v.memberName,v.liKey, v.hierarchyKey)
               table.insert(filterArray,{["hierarchyKey"] =  v.hierarchyKey , ["liKey"] = v.liKey})
            end
         end
      end
   end

   local pass, out_str, code = rest.request( "GET", "config/projects/"..projKey.."/reportConfigs", nil, "Accept:*/*; q=0.01",'application/json;charset=UTF-8')
   local projectReportDetails = Json:decode(out_str)
   for i , v in pairs(reportName) do 
      local res = nil
      for j,w in ipairs(projectReportDetails.items) do
         if w.name == v then
            res = w
            break
         end
      end
      if res == nil then 
         sas.print("%1z".."Report "..v.." Not found . Cancelling Run.")
         return nil
      end
      table.insert(reportIDs, res.id)
   
      -- The body, as a table (will encode as a json string when passing this to rest.request)
      local requestJsonTable={
      ["comment"]=runComment,
      ["ids"] = reportIDs,
      ["filters"] = filterArray
      }
      Debug(Json:encode(requestJsonTable))
      
      local pass, out_str, code = rest.request( 'POST', "reports/"..projKey.."/preview", Json:encode(requestJsonTable), "Accept:*/*; q=0.01",'application/json;charset=UTF-8')
      
      --If requested, wait until the report is done running.
      local wait = 0
      while (true and wait < max_wait) do
         local pass, str = rest.request('GET','config/projects/'..projKey..'/reportConfigs/'..tostring(res.id))
         local status = Json:decode(str).history[1].reportStatus
         Debug("Report status is: ",status)
         if status ~= 'Executing' then break end
         Debug('Checking again in '..tostring(wait_interval)..' seconds ...')
         sas.sleep(wait_interval,1)
         wait = wait + wait_interval
      end   
   end
end 

---Return the attachment id associated with the last run of a report.
-- @param project_id The project id
-- @param report_name the name of the report
-- @return attachment_id The id for the attachment of the last report run. The attachment itself will be a file called 'att_#' where # is this id.
-- @usage
--   local staging = rfw.staging_dir_get()
--   local att_path  = staging:sub(1,staging:find("[/\\]staging")).."attachments/"
-- local att_key = rfw.report_get_most_recent_attachment(project_key,report_name)
-- print("Attachment is: "..att_path..'att_'..tostring(att_key))
function M.report_get_most_recent_attachment( project_id, report_name )
   Trace("report_get_most_recent_attachment", project_id, report_name )
   local pass, json_out = rest.request('GET','config/projects/'..tostring(project_id)..'/reportConfigs')
   if not pass then
      sas.print('%1z '..tostring(json_out))
   end
   local reports = Json:decode(json_out)
   for _,rep in ipairs(reports.items) do
      if rep.reportName == report_name then
        if not rep.history then return end
        return rep.history[1].attachmentKey
     end
   end
   return
end
   
-- *****************************************************
-- *                   #WORKGROUP                      *
-- *****************************************************

--- Work Group Functions
--@section UserWorkgroups
 M.user_workgroups = {}


--- Retrieve a list of the work groups that the current user belongs to.
--  @return a table of work groups for the current user. The Public work group is represented by `""`.
function M.user_workgroups_get(  )
   Trace("user_workgroups_get")
   --  check to see if we already got them
   if M.user_workgroups and M.user_workgroups[1] ~= nil  then return M.user_workgroups end

   -- Get the list of workgroups from the RFG (Risk Work Group) service for this user.
   local rfw_wrkgrps = {}
   local hostname, port, protocol, appName, rwg_url = metadata.get_rest_end_point_dtls("SASRiskWorkGroup")
   rest.set_url(rwg_url)
   local pass,json = rest.request("get","currentUser")
   M.url_set()
   if not pass then return end

   -- Go through these workgroups, if it is associated with RFW, add it and its location to the list
   local wrkgrps = Json:decode(json)
   Debug("Getting workgroup details for user",wrkgrps.name)
   for i,group in ipairs(wrkgrps.groups) do
    for j,solution in ipairs(group.solutions) do
       if solution.twelveByte == 'stresssvrc' then
          rfw_wrkgrps[#rfw_wrkgrps+1] = {name = group.name, location = solution.location}
          break
       end
    end
   end
   Debug(table.tostring(rfw_wrkgrps))

   M.user_workgroups = rfw_wrkgrps
   return M.user_workgroups
end

-- *****************************************************
-- *                     #TASK                         *
-- *****************************************************

--- Workflow Task Functions
--@section Tasks

--- Complete a task in a workflow.
--  @param project_key the numeric identifier for the project that is using the process definition for the workflow.
--  @param task_name the name of the task in the workflow. Task names are case sensitive.
--  @param comment[opt] a string that specifies a comment to add to the completed task.
--  @return `pass`: a Boolean value that indicates whether the task was completed. A value of true indicates success. 

function M.task_complete( project_key, task_name, comment )
   Trace("task_complete", project_key, task_name, comment)
   check_args(project_key, task_name, comment,{["fn"]="task_complete", "number","string",{"string",nil}} )
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return false 
   end
   
   if not comment then comment = " " end

   local process_key = nil
   -- Get the process key for the given the project Key
   local pass, json_out, code = rest.request( "get", "config/projects/")
   local projs = Json:decode(json_out).items
   for i,proj in ipairs(projs) do
      if proj.projectKey == project_key then
         Debug(proj)
         process_key = proj.processKey
         break
      end
   end
   if not process_key then
--       sas.print("%1z Unable to find the project associated with project key")       
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_task_complete_1', 'NOQUOTE')       )
      return false
   end

   -- Get the task ID for 
   local task_id = nil
   pass, json_out, code = rest.request( "get", "tasks/processes/"..process_key)
   local tasks = Json:decode(json_out).tasks
   for i,task in ipairs(tasks) do
      if task.name == task_name then
         Debug(task)
         task_id = task.id
         break
      end
   end

   local body = '{"comment":"'..comment..'","statusKey":2}'
   local header = 'Accept:*/*\nAccept-Encoding:gzip, deflate, sdch'
   local request = 'tasks/processes/'..process_key..'/tasks/'..task_id
   local ct = 'application/json;charset=UTF-8'

   pass, json_out, code = rest.request( "PUT", request, body, header, ct)
   if not pass then
--       sas.print("%1z %s",json_out)
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_task_complete_2', 'NOQUOTE'),json_out)
   end
   return pass
end

--- Retrieves a list of all workflow tasks in a specified project.
--  @param project_id the numeric identifier for the project that contains the workflow. If you have only the project name, you can derive
--  the project ID by using the `project_key_get` function.
--  @return `tasks`: an indexed table that contains the tasks. Each task element in the table includes the following information:
-- 
--  * iterationNum 
--  * description
--  * notificationType
--  * validActions (table)
--  * processKey
--  * users (table)
--  * status (table, with `value` as the only element)
--  * mode
--  * type
--  * script (table). The table contains `key` and `runkey`. If no script is associated with the task, the value for both properties is 0.
--  * id
--  * owner
--  * projectName
--  * dueInDays
--  * name
--  * projectKey
--  * notificationParms
--  * commentCount
function M.task_get_all( project_id )
   Trace("task_get_all", project_id)
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return 
   end
   local process_key = M.project_process_key_get(project_id)
   if not process_key then
      return
   end
   local pass, out_str, code = rest.request("GET", "/tasks/processes/"..process_key.."/alltasks")
   local tasks = Json:decode(out_str)
   return tasks
end

--- Retrieve the details for a task in a specified project.
--  @param project_id the numeric identifier for the project that contains the workflow. If you have only the project name, you can derive
--  the project ID by using the `project_key_get` function.
--  @param task_id the ID of the task for which you want information. Unlike IDs that are system-generated, this ID is specified in the 
--  process definition that is used for the workflow.
--  @return an indexed table that contains the tasks. Each task element in the table includes the following information:
-- 
--  * iterationNum 
--  * description
--  * notificationType
--  * validActions (table)
--  * processKey
--  * users (table)
--  * status (table, with `value` as the only element)
--  * mode
--  * type
--  * script (table). The table contains `key` and `runkey`. If no script is associated with the task, the value for both properties is 0.
--  * id
--  * owner
--  * projectName
--  * dueInDays
--  * name
--  * projectKey
--  * notificationParms
--  * commentCount
function M.task_get( project_id,task_id )
   Trace("task_get", project_id,task_id)
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return 
   end
   local process_key = M.project_process_key_get(project_id)
   if not process_key then
      return
   end

   local tasks = M.task_get_all(project_id)
   for i,v in ipairs(tasks.items) do
      if v.id == task_id then
         return v
      end
   end
   sas.print("%1zUnable to find a task with id %s",task_id)
   return
end

--- Retrieve the current status of a task in a specified project.
--  @param project_id the numeric identifier for the project that contains the workflow. If you have only the project name, you can derive
--  the project ID by using the `project_key_get` function.
--  @param task_id the ID of the task for which you want information. Unlike IDs that are system-generated, this ID is specified in the 
--  process definition that is used for the workflow.
--  @return `status`: a string with the status of the task.
--  @return `status_key`: the numeric identifier of the status.
function M.task_status_get( project_id,task_id )
   Trace("task_status_get", project_id,task_id)
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return 
   end
   local task = M.task_get(project_id,task_id)
   if not task then return end
   return task.status.value, task.status.statusKey
end

--- Set the status of a task.
--  @param project_id the numeric identifier for the project that contains the workflow. If you have only the project name, you can derive
--  the project ID by using the `project_key_get` function.
--  @param task_id the ID of the task whose status you want to set. Unlike IDs that are system-generated, this ID is specified in the 
--  process definition that is used for the workflow.
--  @param status a string with the status you want to set the task to. You can retrieve a list of valid status values by calling the 
--  `task_get` function and examining the entries in the `validActions` table.
--  @param comment[opt] a string that specifies a comment to add to the task if you complete it.
--  @return `pass`: a Boolean value that indicates whether the task status was set. A value of true indicates success.
function M.task_status_set( project_id, task_id, status, comment )
   Trace("task_status_set", project_id, task_id, status, comment)
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return 
   end
   
   if not comment then comment = " " end
   --status = sas.upcase(status)

   local process_key = M.project_process_key_get(project_id)
   if not process_key then return false end
   
   -- Get the task information and check that the status we are setting is valid. If it is, get the status code.
   local task = M.task_get(project_id,task_id)
   if not task then return false end
   local status_key
   local valid_codes = ""
   for i,v in ipairs(task.validActions) do
      Debug(v.statusId,status)   
      if string.upper(v.statusId) == string.upper(status) then
         status_key = v.statusKey
         break
      end
      valid_codes = valid_codes .." "..v.statusId
   end
   if not status_key then
      sas.print("%1zThe status %s is not a valid status for this task. Valid options are: %s",status,valid_codes)
      return false
   end

   local body = '{"comment":"'..comment..'","statusKey":'..status_key..'}'
   Debug(body)
   local header = 'Accept:*/*\nAccept-Encoding:gzip, deflate, sdch'
   local request = 'tasks/processes/'..process_key..'/tasks/'..task_id
   local ct = 'application/json;charset=UTF-8'

   pass, json_out, code = rest.request( "PUT", request, body, header, ct)
   if not pass then
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_task_complete_2', 'NOQUOTE'),json_out)
   end
   return pass
end


--- Wait for a script to finish running after the script is launched from a task.
--  @param projectId the numeric identifier for the project that contains the workflow. If you have only the project name, you can derive
--  the project ID by using the `project_key_get` function.
--  @param taskId the ID of the task that launched the script. Unlike IDs that are system-generated, this ID is specified in the 
--  process definition that is used for the workflow.
--  @param interval the number of seconds to wait before checking the current status for the script. The default value is 5 seconds.
--  @param max_wait the total number of seconds to wait before exiting. The default value is 300 seconds.
--  @return `response`: a string that contains the last status for the script when the `task_wait` function exited. You can use this value
--  to determine whether the task completed or failed.  
function M.task_wait( projectId,taskId,interval,max_wait )
   Trace("task_wait", projectId,taskId,interval,max_wait)
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return 
   end
   local response
   interval=interval or 5
   max_wait = max_wait or 300
   local total_wait = 0
   while total_wait < max_wait do
      response=M.task_get(projectId,taskId)
      if string.upper(response.script['status'])=='EXECUTING' then 
         Debug("Waiting for script to complete...")
         sas.sleep(interval,1)
         total_wait = total_wait + interval
      else 
         break
      end
   end
   return response
end


--- Launch a script from an active task. The system verifies that the task is in progress and that there is a script associated 
--  with the task.
--  @param projectId the numeric identifier for the project that contains the workflow. If you have only the project name, you can derive
--  the project ID by using the `project_key_get` function.
--  @param taskId the ID of the task that launches the script. Unlike IDs that are system-generated, this ID is specified in the 
--  process definition that is used for the workflow.
--  @param scriptParameters a table with name-value pairs of script parameters to override when the script is run. Specify only those parameters 
--  whose default values you want to override.
--  @param waiting[opt] a Boolean value that indicates whether you want to wait for the script to finish running before proceeding. Specify a value of true
--  to wait. The default value is false.
--  @param interval[opt] the number of seconds to wait before checking the current status for the script. The default value is 5 seconds. 
--  This value is used only if the value of `waiting` is true.
--  @param max_wait[opt] the total number of seconds to wait before exiting. The default value is 300 seconds.
--  @param workgroup[opt] The workgroup to run the script under (log will be here). If nil, default workgroup as specified in the process definition is used. 
--  If none is specified there, the default behavior is to use public (note that if none is specified here, the api determines the default behavior).
--  @return a string that contains the last status for the script when the `task_launch_script` function exited. You can use this value
--  to determine whether the task completed or failed.  
--  @usage
--  local resp=task.task_launch_script(47,"102",{{name='Parameter_2', value= "newval"}})
--  print(Json:encode(resp))
function M.task_launch_script( projectId,taskId,scriptParameters,waiting,interval,max_wait,workgroup)
   Trace('task_launch_script', projectId,taskId,scriptParameters,waiting,interval,max_wait,workgroup)
   check_args(projectId,taskId,scriptParameters,waiting,interval,max_wait,workgroup,{["fn"]="task_launch_script", "number","string",{"table","string",'nil'},{"boolean",'nil'},{"number",'nil'},{"number",'nil'},{"string",'nil'} })
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return 
   end
   
   --Allow user to pass a json string for script parameters
   if type(scriptParameters) == 'string' then
      scriptParameters=json:decode(scriptParameters)
   end
   --Getting task details
   local response , scriptParamTab
   local taskDetail= M.task_get(projectId,taskId)
   waiting = waiting or false
   Debug(Json:encode(taskDetail))

   --Checking if task has script and it is active (in progress).
   if string.upper(taskDetail.id) == string.upper(taskId) and string.upper(taskDetail.status['value']) == 'ACTIVE' and taskDetail.impl=='script.run' and taskDetail.script['key'] > 0 then 
      local pass, out_string ,code = rest.request("GET", "/scripts/"..taskDetail.script['key'])
      scriptResp = Json:decode(out_string)
      scriptResp['code']=nil
      scriptResp['updated']=nil
      scriptResp['created']=nil

      -- If there are any script parameters to update, get the defaults and then update them
      if scriptParameters then 
          for j,k in pairs(scriptParameters) do 
            for i,v in pairs(scriptResp.defaultParameters)do
               if string.upper(k.name) == string.upper(v.name) then 
                  Debug("parameter name",k.name,"is using value",k.value,"instead of",v.value)
                  v.value=k.value
                  break
               end
            end
         end
      end
      local url = "/tasks/processes/"..taskDetail['processKey'].."/tasks/"..taskId.."/run"
      if workgroup then
         url = url .. '?workGroupName='..utils.urlencode(workgroup)
      end
      
      local  pass, out_string ,code = rest.request( 'POST', url, Json:encode(scriptResp), "Accept:*/*; q=0.01",'application/json;charset=UTF-8')
      if not pass then 
         sas.print("%1z"..tostring(code),out_string)
        return  
      else 
         response=Json:decode(out_string)
         if waiting then 
            response=M.task_wait(projectId,taskId,interval,max_wait)
            return response
         else 
            return response
         end
      end
   else 
      --If task not active or not found or doesn't have script
      sas.print("%1z".."Task ID:: "..taskId.." not found or Task is not active or task implementation is not script.run")
      return 
   end 
end

--- Attach a file to a task
-- @param 'projectId' Id of the project
-- @param 'taskId' Id of the project task where the attachment must be loaded
-- @param 'file' The full path to the file that must be uploaded 
-- @param 'comment' (Optional) Comment to be posted together with the attachment
-- @return 'pass,out_str' - a boolean indicating success and the output body from the http call
function M.task_load_attachment( projectId, taskId, file, comment )
   Trace('task_load_attachment', projectId, taskId, file, comment)
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return 
   end
   comment = comment or ""

   -- Extract the file name
   local filename = sas.scan(file,-1,"\\/")
  
   -- Build the request body (Multipart)
   local topText =    '------WebKitFormBoundarynL7rIPgAavvpQxNG\r\n'
   topText = topText.. 'Content-Disposition: form-data; name="file"; filename="'..filename..'"\r\n'
   topText = topText.. 'Content-Type: application/octet-stream\r\n\r\n'

   local bottomText =    '\r\n------WebKitFormBoundarynL7rIPgAavvpQxNG\r\n'
   bottomText = bottomText.. 'Content-Disposition: form-data; name="_charset_"\r\n\r\n'
   bottomText = bottomText.. 'UTF-8'
   
   bottomText = bottomText.. '\r\n------WebKitFormBoundarynL7rIPgAavvpQxNG\r\n'
   bottomText = bottomText.. 'Content-Disposition: form-data; name="name"\r\n\r\n'
   bottomText = bottomText.. filename
   
   bottomText = bottomText.. '\r\n------WebKitFormBoundarynL7rIPgAavvpQxNG\r\n'
   bottomText = bottomText.. 'Content-Disposition: form-data; name="description"\r\n\r\n'
   bottomText = bottomText.. comment..'\r\n\r\n'
   
   bottomText = bottomText.. '------WebKitFormBoundarynL7rIPgAavvpQxNG--'

   -- Create temporary file in work to store the request body
   local fbodyPath = sas.pathname("work") .. "/" .. "fbody.txt"
   local in_logical = sasxx.assign(fbodyPath, "fbody")
   
   local xl_logical = sasxx.assign(file,"xl")
   
   local body = topText
   local handle = xl_logical:open('r-o','fc')
   if not handle then
--       sas.print("%1z Could not open "..file..". Check that it exists and is readable")
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_process_create_1', 'NOQUOTE'),file)
   end   
   for line in handle:lines() do
      body=body..line
   end
   handle:close()
   body = body..bottomText
   handle = in_logical:open('w+o',nil,string.len(body))
   handle:write(body)
   handle:close()
   xl_logical:deassign()   
   
   
   -- Send the request
   local url = "attachments/project/" .. projectId .. "/task/" .. taskId
   local pass, out_str, code = rest.request("POST", url, {[1] = 'fbody'}, 'accept: */*', 'multipart/form-data; boundary=----WebKitFormBoundarynL7rIPgAavvpQxNG', nil)
   
   sas.fdelete("fbody")
   sas.filename("fbody")

   return pass,out_str
end

--- Start a new iteration for a process
-- @param 'project_id' The id for the project
-- @param 'task_id' The id of the task to start the new iteration at
-- @return 'out_str,pass' - the output body from the http call and a boolean indicating success
function M.task_new_iteration( project_id, task_id)
   Trace("task_new_iteration", project_id, task_id)
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return 
   end
   
   local url = 'config/projects/'..tostring(project_id)..'/nextiteration'
   local header = 'Accept: */*\nAccept-Encoding: gzip, deflate'
   local body = '------WebKitFormBoundaryAbtyT2AHu9uXxuMS\r\nContent-Disposition: form-data; name="taskId"\r\n\r\n'..tostring(task_id)..'\r\n------WebKitFormBoundaryAbtyT2AHu9uXxuMS--'
   local pass, out_str, code = rest.request("POST", url, body, header, 'multipart/form-data; boundary=----WebKitFormBoundaryAbtyT2AHu9uXxuMS')

   return pass,out_str
end

-- *****************************************************
-- *                  #DETAIL DATA                     *
-- *****************************************************

--- Detail Data Functions
--@section DetailData
 
--- Retrieve the directory path for detail data for a specified project. You can identify the project with either the project name
--  or the project ID. This function uses the `project_path_get` function to enable you to specify the project name instead of a key.
--  @param projectId the numeric identifier for the project that contains the detail data that you are interested in.
--  @param workgroup the name of the work group that the detail data is associated with. Only users who belong to the work group 
--  can access the detail data.
--  @param detailKind the type of detail data.
--  @return a Boolean value that indicates whether the directory path was located. A value of true indicates success.
--  @return `detailLoc`: the full path to the detail data directory. 
--  @return `project_key`: the key value that identifies the project.
function M.detail_data_loc_get( projectId, workgroup, detailKind )
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return 
   end
       Trace("detail_data_loc_get",projectId, workgroup, detailKind)
       local appDataRoot 
       if M.testing then
          appDataRoot = root_dir or M.root_dir_get()
          appDataRoot = sas.substr(appDataRoot,1,sas.findc(appDataRoot,'/\\','b')-1)
       end
       local detailLoc = M.project_path_get( projectId, workgroup, detailKind,appDataRoot)
       Debug(detailLoc)
       if  detailLoc == nil then
          return false
       end  
       -- check for access
       local testdir = sas.filename("__test", detailLoc .."/_testwrite")
       if testdir > 0 then
         Debug("Failed to open fileref")
         return false 
       end
       local fid = sas.fopen("__test", "O")
       if fid ~= 0 then 
          sas.fclose(fid)
          sas.fdelete("__test")
          sas.filename("__test")
          return true, detailLoc, projectId
       end
       Debug("Failed to open test file")
       return false
end

--- Retrieve the directory path for detail data for a specified project and work group. If the directories in the path do not exist,
--  they are created.
--  @param project_id the numeric identifier for the project.
--  @param workgroup the name of the work group that the detail data is associated with. Only users who belong to the work group 
--  can access the detail data.
--  @param detailKind the type of detail data.
--  @param appdata_root[opt] a parameter that is used internally for testing.
--  @return `root_path`: the full path to the detail data directory for this work group and project.
function M.project_path_get(  project_id, workgroup, detailKind, appdata_root )
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return 
   end
    local workgroups = M.user_workgroups_get()
    local root_path
    -- check to see if work_groups exists
   if workgroup ~= '' then 
      for i,grp in ipairs(workgroups) do
         if sas.lowcase(workgroup) == sas.lowcase(grp.name) then
            root_path = grp.location
            break
         end
      end

      if sas.fileexist(root_path) == 0 then 
          -- If testing, make workgroup fodlers if they are not there, otherwise this is an error
          if M.testing then
              root_path = sas.dcreate( "work_groups" , appdata_root )
              print("Creating testing path ",root_path)
              if string.len(root_path) < 5 then
--                  sas.print("%1zCannot create for testing work group directory in %s/work_groups.", appdata_root)
                 sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_project_path_get_1', 'NOQUOTE'), appdata_root)
                return nil 
              end          
              root_path = sas.dcreate( workgroup , old_path )
              print("Creating workgroup testing path ",root_path)
              if string.len(root_path) < 5 then
--                  sas.print("%1zCannot create for testing work group directory in %s.", old_path .. "/" .. workgroup)
                 sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_project_path_get_3', 'NOQUOTE'), old_path .. "/" .. workgroup)
                return nil 
              end
          else
--              sas.print("%2zCannot find work group directory in %s. User may not be a member of the workgroup %s.", root_path, workgroup)
             sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_project_path_get_4', 'NOQUOTE'), root_path, workgroup)
             return nil 
          end
      end
    end
    
    -- check to see if detail_data exists
    local old_path = root_path..'/';
    root_path =  root_path .. '/detail_data';
   
    if sas.fileexist(root_path) == 0 then 
        if M.testing then           
            root_path = sas.dcreate( "detail_data" , old_path )
            print("Creating detail testing path ",root_path)
            if string.len(root_path) < 5 then
--                sas.print("%1zCannot create for testing detail_data directory in %s.", old_path .. "/" .. workgroup)
               sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_project_path_get_5', 'NOQUOTE'), old_path .. "/" .. workgroup)
              return nil 
            end
        else
--            sas.print("%2zCannot find detail_data directory in %s. User may not be a member of the workgroup %s.", root_path, workgroup)
           sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_project_path_get_6', 'NOQUOTE'), root_path, workgroup)
           return nil 
        end
    end
    
        -- check to see if project dir exists
    local old_path = root_path..'/';
    root_path =  root_path .. '/' .. project_id;
   
    if sas.fileexist(root_path) == 0 then 
        root_path = sas.dcreate( tostring(project_id) , old_path )
        
        if string.len(root_path) < 5 then
--            sas.print("%2zCannot create or find project directory in %s. User may not be a member of the workgroup %s.", root_path, tostring(project_id))
           sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_project_path_get_7', 'NOQUOTE'), root_path, tostring(project_id))
           return nil 
        end
    end
    
           -- check to see if detailKind dir exists
    local old_path = root_path..'/';
    root_path =  root_path .. '/' .. detailKind;
   
    if sas.fileexist(root_path) == 0 then 
        root_path = sas.dcreate( detailKind , old_path )
        
        if string.len(root_path) < 5 then
--            sas.print("%2zCannot create or find detailKind directory in %s. User may not be a member of the workgroup %s.", root_path, detailKind)
           sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_project_path_get_8', 'NOQUOTE'), root_path, detailKind)
           return nil 
        end
    end
      
    return root_path
end

--- Retrieve the detail data for a specified project.
--  @param projectId the numeric identifier for the project.
--  @param detailKind the SAS name that indicates the type of detail data. Data with the same variable-to-hierarchy mapping should be the same kind. 
--  @param output_ds the name of the output data set (for example, WORK.FOO or SASUSER.MINE).
--  @param stage[opt] the name of the data set that corresponds to the processing stage of the detail data. The default value is the data set
--  for the last stage. Stage names must conform to the naming rules for SAS data sets.
--  @param where_cls[opt] a WHERE clause to apply to the selected data.
--  @param details[opt] a Boolean value that indicates whether to join data with details. This parameter applies only if data was loaded in 
--  separate parts. The default value is false (no join).
--  @param workgroups[opt] a table that contains a list of work group to retrieve detail data for. The default behavior is to retrieve data
--  for all work groups that the current user belongs to.
--  @return a Boolean value that indicates whether the detail data was retrieved. A value of true indicates success.
--  @return `workgroups`: a list of the work groups for which detail data was retrieved.
function M.detail_data_get( projectId, detailKind, output_ds, stage, where_cls, details, workgroups )
       Trace("detail_data_get",projectId, detailKind, output_ds, stage, where_cls, details, workgroups)
       check_args( projectId, detailKind, output_ds, stage, where_cls, details, workgroups,
                 {["fn"]="detail_data_get", "number", "string", "string", {"string", 'nil'}, {"string", 'nil'}, {"boolean", 'nil'}, {"table", 'nil'}} )
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return 
   end
       if not workgroups then
          workgroups = {}
          M.user_workgroups_get()
          for i,grp in ipairs(M.user_workgroups) do
             workgroups[#workgroups+1] = grp.name
          end
       end
       local wrkgrps_usd = {}
       local loadid
       
       stage = stage or "final"
       Debug("Stage ", stage)     
       for i,wgrp in ipairs(workgroups) do
         --  find corresponding detail data location
         -- Get the project key if necessary
         local pass, detailLoc = M.detail_data_loc_get( projectId, wgrp, detailKind )
         
         if  pass then         
           sas.submit([[
                libname _det "@detailLoc@" access=readonly ;
                ]],nil,nil,4)
           -- Four cases
           -- 1. No where no join
           -- 2. No where yes join
           -- 3. Yes where and yes join 
           ---4. Yes where and no join   
       
           details = sas.exists('_det.join') and details and alloc.var_exists('_det.initial', '_link')
           
            -- Case 1
           local get_data 
           if not where_cls and not details then
               get_data = "create table @out_ds@(compress=yes ) as select * from _det.@stage@;"
           elseif not where_cls and details then  -- Case 2
               get_data = "create table @out_ds@(compress=yes ) as select * from _det.@stage@ a inner join _det.join b on a._link = b._link;"
           elseif where_cls and details then  -- Case 3
               get_data = [[create table tmp_ds(compress=yes ) as select * from _det.@stage@ where @where_cls@;
                            create table @out_ds@(compress=yes ) as select * from tmp_ds a inner join _det.join b on a._link = b._link ;
                          ]]
           else
               get_data = "create table @out_ds@(compress=yes ) as select * from _det.@stage@ where @where_cls@;"        
           end             
                
           local out_ds = output_ds 
           if #wrkgrps_usd > 0 then out_ds = 'temp_ds' end
             
           local final_get_data = string.resolve(get_data)
             
           if sas.exists('_det.'..stage) or sas.exist('_det.'..stage, 'VIEW') == 1 then 
             if #wrkgrps_usd == 0 then
               sas.submit([[                                      
                  proc sql noprint nowarn;                                        
                      @final_get_data@
                  quit;                                                      
                  
                  libname _det;
               ]],nil,nil,4)  -- for some reason this sql sets the syserr to 4
             else
               sas.submit([[                 
                  proc sql noprint nowarn;
                      @final_get_data@
                  quit;           
                                               
                  proc append base=@output_ds@(compress=yes) data=temp_ds force;
                  run;
                  
                  libname _det;
               ]],nil,nil,4)
             end
             wrkgrps_usd[#wrkgrps_usd+1] = wgrp
           end
         end
       end
       if #wrkgrps_usd > 0 then
          return true, wrkgrps_usd
       else
          return false
      end
end

local function check_for_required_vars(inputds, ccvs, workgroup, entity )
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return 
   end
       local bucketCode, input_ds
       local rename = ""
       local req_vars = string.split(ccvs," ")
       req_vars[#req_vars+1] = "value"

       if type(inputds) == "table" then
          input_ds = inputds.values
          req_vars[#req_vars+1] = "_link"
          if not alloc.var_exists(inputds.join,"_link") then
--               sas.print("%1zMissing variable %s in data set %s.", v, input_ds)
              sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_detail_data_put_2', 'NOQUOTE'), "_link", inputds.join)
              return false
           end
       else
          input_ds = inputds
       end
       
       if not workgroup then   --Require workgroup if not passed as an arg
          req_vars[#req_vars+1] = "workgroup"
       end
       if entity == "" then   --Require entity if not passed as an arg
          req_vars[#req_vars+1] = "entity"
       end
       local success = true
       for i,v in ipairs(req_vars) do
           if not alloc.var_exists(input_ds,v) then
--               sas.print("%1zMissing variable %s in data set %s.", v, input_ds)
              sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_detail_data_put_2', 'NOQUOTE'), v, input_ds)
              success = false
           end
       end
       if not success then return false end
       -- allow either "date" or "evaldate"
       if not alloc.var_exists(input_ds,"evaldate") then
          if not alloc.var_exists(input_ds,"_date_") then
--              sas.print("%1zMissing variable 'date' or 'evaldate' in data set %s.", input_ds)
             sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_detail_data_put_3', 'NOQUOTE'), input_ds)
             return false
          end
          rename = "evaldate = _date_;"
       end
       
       return true, rename
end

local function bucket_code(interval)
       local bucketCode
       
       bucketCode = string.resolve([[put( intnx("@interval@", evaldate, 0), date.) ]])
       
       return(bucketCode)
end

--- Load the detail data for a specified project.
-- 
--  You can load data from a single input data set, or you can use two input data sets to separate numbers and classifications 
--  from the rest of the details. Separate data sets can be important when you have either multiple numbers or multiple
--  horizons per detail data.
--
--  The following example shows a SAS program that creates separate input data sets (`_DET_.VALUES` and `_DET_.JOINDS`)
--  from an initial data set that has multiple numbers to store in SAS Risk and Finance Workbench.
-- 
--  [//]: # (Markdown doesn't support comments. However, this syntax works like a comment, and won't appear in any output.)
--  [//]: # (Reminder: To ensure that the code example displays correctly after Markdown transform, code lines must be indented 4 spaces.)
--    data _det_.join(drop=&ccvs evaldate value compress=yes) _det_.values(keep= &ccvs evaldate value  _link compress=yes);
--       set _det_.allprice;
--       length _link $10;
--       _link = "L" || compress(put(_N_, z9.));
--       value = ECL;
--       MovementType = "";
--       output _det_.values; output _det_.join;
--       MovementType = "Balance_Beginning of Period";
--       value = UPB;
--       output  _det_.values;
--       MovementType = "Prepayments and run_offs";
--       value = PREPAY_AMT_LIFETIME; 
--       output  _det_.values;
--    run;
--  
--  Notes:
--
--  * The _LINK variable is required and is used to connect the two data sets.
--  * The way in which the _LINK variable is created ensures that when the two data sets are indexed and then rejoined (to support drilling
--    to details), the operation is performed sequentially, because the join key is generated in alphabetic order.
--    It is possible to use INSTID for the _LINK variable, but the performance of a drill-down operation will be slower.
--  * Specify the join data only on the initial load.
--  @param projectId the project for which you are loading detail data. You can specify the project name as a string, or you can specify
--  the project ID as a number.
--  @param input_ds the name of the input data set. The data set must have a column named WORKGROUP. If you are using separate data sets for input,
--  you can specify a table for the `input_ds` parameter. In this case, the table should contain two entries (for example, 
--  `{values="work.input", join="work.joinds"}`).
--  @param stage[opt] the name of the data set that corresponds to the processing stage of the detail data. The default value is `initial`.
--  Stage names must conform to the naming rules for SAS data sets.
--  @param workgroup[opt] the name of the work group that this data is to be loaded for. If the value is `nil`, the input data must have a 
--  column named WORKGROUP.
--  @param detailKind the SAS name that indicates the type of detail data. Data with the same variable-to-hierarchy mapping should be the same kind. This 
--  parameter is required for the initial load.
--  @param libRefDef[opt] the logical name that identifies the library where the input data set is located. This parameter is required if the input data set
--  is not in a location that is known to SAS Risk and Finance Workbench. The value should be a valid SAS LIBNAME statement.
--  @param entity[opt] the name of the entity that this data is associated with. If there is an ENTITY column in the data, this parameter is ignored.
--  @param scenario[opt] the name of the scenario that this data is associated with. If there is a SCENARIO_NAME column in the data, this parameter is ignored.
--  @param mgversion[opt] the version of the model group to use for the load. The default value is `""`.
--  @return a Boolean value that indicates whether the detail data was loaded. A value of true indicates success.
--  @return `loadId`: a numeric identifier that identifies a given transfer of detail data into the system.
function M.detail_data_put( projectId, input_ds, stage, workgroup, detailKind, libRefDef,  entity, scenario, mgversion)
       Trace("detail_data_put",projectId, input_ds, stage, workgroup, detailKind, libRefDef,  entity, scenario, mgversion) 
       check_args( projectId, input_ds, stage, workgroup, detailKind, libRefDef, entity, scenario, mgversion,
                 {["fn"]="detail_data_put", "number", {"string", 'table'}, {"string", 'nil'}, {"string", 'nil'}, "string", {"string", 'nil'}, {"string", 'nil'},{"string", 'nil'},{'string','nil'}} )
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return 
   end                 
       -- first check for existence of the input data
       mgversion = mgversion or ""
       local in_ds = input_ds
       if type(input_ds) == table then in_ds = input_ds.values end
       if (type(input_ds) == "string" and not sas.exists(input_ds, 'DATA')) or 
          (type(input_ds) == "table" and not (sas.exists(input_ds.values, 'DATA') and sas.exists(input_ds.join, 'DATA'))) then
--            sas.print("%1zDetail data, %s,doesn't exist.", input_ds)
           sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_detail_data_put_1', 'NOQUOTE'), in_ds)
          return false
       end
       
       entity = entity or ""
       scenario = scenario or ""
       local interval 
       
       -- look for a Load_Details_@stage@ model to create model group runs for loading data
       -- Get HierarchyMapping and ccvs
       local ccvs, hm, hierarchyMapping
       if M.testing then
          ccvs = sas.symget("ccvs")
          hierarchyMapping = sas.symget("hierarchyMapping")
          interval = sas.symget("interval")
       else
          ccvs, hm, hierarchyMapping, interval = M.ccvs_hierarchy_get( projectId, "Load_Details_" .. detailKind, mgversion)
       end
       if not ccvs then 
          return
       end
         
       -- Then check for required variables
       local success, rename = check_for_required_vars( input_ds, ccvs, workgroup, entity) 
       if not success then return false end
       
       stage = stage or "initial"

       -- get the list of workgroups from the input data
       local workgroups = {}
       if not workgroup then
         local wrkgrps1 = M.user_workgroups_get() 
         sas.submit([[
            proc sql noprint;
             select distinct workgroup into
              :workgroups separated by ',' from @in_ds@ ;
            quit;
          ]])
         local wrkgrps2 = string.split(sas.symget("workgroups"), ',') 
         if #wrkgrps2 == 0 then
            wrkgrps2 = {""}
         end
         -- 'and' the two lists  
         for i,g1 in ipairs(wrkgrps1) do 
             for j,g2 in ipairs(wrkgrps2) do
                 if sas.lowcase(g1.name) == sas.lowcase(g2) then
                    workgroups[#workgroups+1] = g1.name
                 end
             end
         end
       else
         workgroups = {workgroup}
       end
       
       local wrkgroups_used = {}
       local input_dataset = input_ds
       if type(input_ds) == "table" then input_dataset = input_ds.values end
       
       local bucketCode = bucket_code(interval)
       local indexccvs = "entity date_bucket scenario_name ".. ccvs
       if alloc.var_exists(input_dataset,"_link") then
          indexccvs = '_link ' .. indexccvs
       end
       
       local pass, detailLoc
       local libdecl, destdecl, seldecl, viewsdecl, modifydecl = "","","","",""
       for i,wgrp in ipairs(workgroups) do
         --  find corresponding detail data location
         -- Get the project key if necessary
         pass, detailLoc = M.detail_data_loc_get( projectId, wgrp, detailKind )
         if pass then 
           wrkgroups_used[#wrkgroups_used+1] = wgrp
           libdecl = libdecl .. "libname _det"..i.." '".. detailLoc.."' compress=yes;\n"
           destdecl = destdecl .. "_det"..i.."."..stage .. " "
           seldecl = seldecl .. 'if upcase(workgroup) = upcase("' ..wgrp..'") then output _det'..i.."."..stage .. ";\n"
           viewsdecl = viewsdecl.. 'create view _det'..i..'.final as select * from _det.'..stage .. ";\n"
           modifydecl = modifydecl..'proc datasets noprint library=_det'..i..";\n modify "..stage..";\n index create "..indexccvs..";\n run;"
         end
       end
       
       -- put data all at once 
       local setworkgroup = ""  
       if workgroup then
          setworkgroup =  "workgroup = '" .. workgroup.."';\n" 
       end         
       sas.submit([[@libdecl@
                    data @destdecl@; length workgroup $64 entity $1024 scenario_name $64;
                        set @input_dataset@;                                         
                         
                         @rename@
                         date_bucket = @bucketCode@;
                         @setworkgroup@ 
                         if entity = "" then
                            entity = "@entity@";
                         if scenario_name = "" then
                            scenario_name = "@scenario@";  
                            
                         @seldecl@      
                     run;                            
                        
                     proc sql noprint;
                        @viewsdecl@;
                     quit;
                     
                     @modifydecl@
                     
                   ]],nil,nil,4)
           
      -- handle join data
      if stage == "initial" and type(input_ds) == "table" then
         local libdecl, createdecl, modifydecl = "","",""
         for i,wgrp in ipairs(wrkgroups_used) do
           --  find corresponding detail data location
           pass, detailLoc = M.detail_data_loc_get( projectId, wgrp, detailKind )
           if pass then 
             libdecl = libdecl .. "libname _det"..i.." '".. detailLoc.."' compress=yes;\n"
             createdecl = createdecl .. "create table _det"..i..".join as select b.* from _det"..i..".initial a inner join " .. input_ds.join .. " b on a._link = b._link;\n" 
             modifydecl = modifydecl..'proc datasets noprint library=_det'..i..";\n modify join;\n index create _link;\n run;"
           end
         end
         sas.submit([[@libdecl@
              proc sql noprint nowarn; 
              @createdecl@
              quit;
              
              @modifydecl@
         ]])
      end
           
      
      local loadid
      if #wrkgroups_used > 0 then
         detailKind = detailKind or "default"
         if stage == "initial" then            
            pass, loadid =  M.detail_metadata_new(wrkgroups_used,projectId,ccvs, hierarchyMapping, detailKind,libRefDef, interval)
         
         else
            pass, loadid = M.detail_metadata_update(wrkgroups_used,projectId, detailKind, stage) 
         end
         
         if not pass then
--            sas.print("%1zFailed to update metadata about details.")
           sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_detail_data_put_4', 'NOQUOTE'))
           return false
         end
      else
--           sas.print("%1zDetail data doesn't contain any workgroups the user has permission to update.")
          sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_detail_data_put_5', 'NOQUOTE'))
          return false
      end
      
      -- Create the model run groups for loading the aggregate data
      local scenarios, entities
      if stage == "initial" then
      
         if scenario == "" then 
           if alloc.var_exists(input_dataset,"scenario_name") then
             sas.submit([[
                   proc sql noprint;
                              select distinct scenario_name into
                              :scn_names separated by ',' from @input_dataset@
                              ;
                           quit;
             ]])
             scenarios = string.split(sas.symget("scn_names"), ',')
             if #scenarios == 0 then scenarios = nil end
           else
             scenarios = nil
           end
         else
             scenarios = {scenario}
         end
         
         if entity == "" then            
             sas.submit([[
                   proc sql noprint;
                              select distinct entity into
                              :ent_names separated by ',' from @input_dataset@
                              ;
                           quit;
             ]])
             entities = string.split(sas.symget("ent_names"), ',')
           
         else
             entities = {entity}
         end
             
         local pass, msg = M.model_group_agg_load_setup(projectId,detailKind,entities,scenarios)

         if not pass then
--            sas.print("%1z%s.", msg)
           sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_detail_data_put_6', 'NOQUOTE'), msg)
           return false
         end
      end
      
      return true, loadid
end

--- Summarize the detail data for a specified project. This function runs the HPSUMMARY procedure.
--  @param projectId the numeric identifier for the project that contains the detail data.
--  @param output_ds the name of the output data set that will contain the summarized data.
--  @param detailKind the SAS name that indicates the type of detail data. Data with the same variable-to-hierarchy mapping should be the same kind. 
--  @param stage[opt] the name of the data set that corresponds to the processing stage of the detail data. The default value is the data set
--  for the last stage. Stage names must conform to the naming rules for SAS data sets.
--  @return a Boolean value that indicates whether the detail data was summarized. A value of true indicates success.
--  @return `wg`: a table that contains a list of the work groups for which data was summarized.
--  @return `loadId`: the numeric identifier for the specific load operation that was used to load the detail data.
--  @return `ccvs`: a string that contains a list of the cross-classification variables, separated by spaces.
--  @return `last_stage`: the last processing stage of the detail data.
--  @return `hierarchyMapping`: a table that maps hierarchy level names to cross-classification variables.
function M.detail_data_sum( projectId, output_ds, detailKind, stage )
   Trace("detail_data_sum", projectId, output_ds, detailKind, stage )
   check_args( projectId,  output_ds, detailKind, stage, 
                 {["fn"]="detail_data_sum", "number", "string", "string", {"string", 'nil'}} )
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return 
   end                 
   local workgroups = {}
   M.user_workgroups_get()
   for i,grp in ipairs(M.user_workgroups) do
      workgroups[#workgroups+1]=grp.name
   end

   local wg = {}
   local pass, loadid, ccvs, last_stage, detailLoc, hierarchyMapping
   local scn_name
   local found_data=false
   
   for n, grp in ipairs(workgroups) do
        --  find corresponding detail data location
       -- we will get an error if a workgroup is in the list but the user doesn't have access
       pass, detailLoc = M.detail_data_loc_get( projectId, grp, detailKind)

       if  pass then
                 
          -- Only need to look this info up once
          if not found_data then 
            pass, loadid, ccvs, last_stage, hierarchyMapping = M.detail_metadata_get(grp, projectId, detailKind)           
            if not pass then
               print("ERROR: Detail metadata missing for ", detailKind)
               return false
            end
            if loadid then 
              stage = stage or "final"
              Debug("Stage " , stage)
              hierarchyMapping = Json:decode(hierarchyMapping)
              found_data = true
            end
          end
          
          sas.submit([[ libname _det "@detailLoc@" access=readonly ;]])
          if found_data and sas.exists('_det.initial') then
          
            wg[#wg+1] = grp
            -- check scenario_name to see if it is blank
            sas.submit([[ 
                          data _NULL_; set _det.@stage@(obs=1); call symput("scenario_name", compress(scenario_name));
                          run;
                        ]])
           
           if sas.symget("scenario_name") == "" then
              scn_name = "";
           else
              scn_name = "Scenario_name"
           end        
  
            -- Create new summary data set
           sas.submit([[
                /* summarize the data */
                  proc hpsummary data=_det.@stage@ nway ;
                  class @ccvs@ date_bucket entity @scn_name@;
                  vars value ;
                  output sum(value)=value  out=sum_temp(drop=_type_ );
                  run;    
                 ]])
                 
            -- Accumulate the summary data sets across workgroups
           if #wg==1 then
              sas.submit([[data work._summary; set sum_temp;length workgroup $64;workgroup="@grp@";run;]])
           else
               sas.submit([[
               data work.sum_temp; set sum_temp;length workgroup $64;workgroup="@grp@";run;
               proc append base=work._summary data=sum_temp force;run;]], nil,nil,4)
           end
         end       
      end
    end
    if #wg == 0 then return end
    
    sas.submit([[
     proc hpsummary data=work._summary(rename=_freq_=count_) nway;
        class @ccvs@ date_bucket entity @scn_name@;
        vars value count_ ;
        output sum(value)=value sum(count_)=count_  out=work.summary(drop=_type_ _freq_);
        run;   
        data work.summary; set work.summary;  rename count_=_freq_ ;
        run;       
       ]])          
       
    return true, wg, loadid, ccvs, last_stage, hierarchyMapping
       
end

--- Copy detail data from final for a specified processing stage.
--  @param projectId the numeric identifier for the project that contains the detail data.
--  @param stage the name of the data set that corresponds to the processing stage of the detail data to copy.
--  @param detailKind the SAS name that indicates the type of detail data to copy.
--  @return a Boolean value that indicates whether the detail data was copied. A value of true indicates success.
function M.detail_data_copy( projectId, stage, detailKind)
--TODO: get the value of "final" -- latest stage
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return 
   end
   local workgroups = M.user_workgroups_get()
   for n, grp in ipairs(workgroups) do
       --  find corresponding detail data location
       -- Get the project key if necessary
       local pass, detailLoc = M.detail_data_loc_get( projectId, grp, detailKind )
       if  pass then        
            sas.submit([[ libname _det "@detailLoc@" compress=yes;
                    data _det.@stage@; set _det.final; run;]])             
       end
    end
    return true
end

--- Create a comma-separated values (CSV) file from the detail data. 
--  @param projectId the numeric identifier for the project that contains the detail data.
--  @param detailDataKind the SAS name that indicates the type of detail data.
--  @param location a table that contains name-value pairs of hierarchy names and line-item keys.
--  @param ccvs_hier_map a table that maps hierarchy level names to cross-classification variables. 
--  Format: `hierarchy name = table {level name = ccv}`.
--  @param destination the full path to use for the generated output file, including file name and extension.
--  @return A Boolean value that indicates whether the detail data was summarized. A value of true indicates success.
--  @usage
--  local location = { region_state_dim = { region="SouthWest", state="AZ" }, loanclass={loantype="Fixed Rate"}} 
--  local ccvs_hier_map = { region_state_dim = { region = "region", state="state"}, loanclass = { loanType = "loantype"}}
--  local pass = rfw.detail_data_to_csv(1, workGrps, location, ccvs_hier_map, "c:\\temp\\loandetails.xlsx" )
function M.detail_data_to_csv( projectId, detailDataKind, location, ccvs_hier_map, destination)
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return 
   end
    -- Make a where clause
    -- first get date bucket
    sas.submit([[
       data _NULL; 
          evaldate = input( "@startDate@", anydtdte11. );
          date_bucket = put( intnx("@interval@", evaldate, 0), date.);
          call symput("date_bucket", trim(date_bucket));
       run;
    ]], {startDate = location.periodStartDate, interval = location.periodType })
    local date_bucket = sas.symget("date_bucket")
    local where_cls
    if location.scenarioName == "" then
      where_cls = "date_bucket = '" .. date_bucket .."'"
    else
      where_cls = "scenario_name = '".. location.scenarioName.. "' and date_bucket = '" .. date_bucket .."'"
    end
    local entname = sas.symget("sst_entity_hierarchy_name")
    if not entname then entname = "Entity" end
    for i, h in ipairs(location.cells) do
       local ccv, skip
       
       for j, lv in ipairs(h.levels) do
          ccv = ccvs_hier_map[h.hierarchy].classificationVarNames[j] 
          if ccv == nil then 
--              sas.print("%1zCan't map ccvs to hierarchy %s.", h.hierarchy)
             sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_detail_data_to_csv_1', 'NOQUOTE'), h.hierarchy)
             return false
          end
           -- special case for entity
          skip = false
          if h.hierarchy == entname and j < #h.levels then skip = true end
          if lv.member ~= "" and not skip then      
            local esc_str = string.gsub(lv.member, "'","''") 
            where_cls = where_cls .. " and " .. ccv .. " = '" .. esc_str .."'"
          end
       end
    
    end
    Debug(where_cls)
    
    -- get the data
    M.detail_data_get( projectId, detailDataKind, "work.csv", nil, where_cls, true )
    
    -- Create the csv
    sas.submit([[
      proc export 
          data=work.csv 
          dbms=csv 
          outfile="@destination@" 
          
          replace;
      run;
    ]])
    
    return true
end

--- Create a new load ID. The load ID is used as metadata to tag the detail data in SAS Risk and Finance Workbench. 
--  The ID identifies a given transfer of detail data into the system. 
--  @param workgroups a table that contains a list of the work groups that the current user belongs to.
--  @param projectId the numeric identifier for the project in which the data is added.
--  @param ccvs a string that contains a list of the cross-classification variables, separated by spaces.
--  @param hierarchyMapping a JSON string that maps hierarchy level names to cross-classification variables.
--  For example: `{'loanclass': {'loanType':'loantype'}, 'region_state_dim': {'region':'region', 'state':'state'}}`
--  @param DetailKind  the SAS name that indicates the type of detail data. Data with the same variable-to-hierarchy 
--  mapping should be the same kind. 
--  @param librefDef[opt] the logical name that identifies the library where the detail data set is located. 
--  This parameter is required if the input data set is not in a location that is known to SAS Risk and Finance Workbench. 
--  The value should be a valid SAS LIBNAME statement.
--  @param interval[opt] the time period that is used by the detail data. The following values are supported: YEAR, SEMIYEAR, 
--  QUARTER, MONTH, WEEK. The default value is MONTH.
--  @return A Boolean value that indicates whether the load ID was created. A value of true indicates success.
--  @return `loadId`: the load ID that was generated.
function M.detail_metadata_new(workgroups, projectId, ccvs, hierarchyMapping, DetailKind, librefDef, interval )
  Trace("detail_metadata_new", workgroups, projectId, ccvs, hierarchyMapping, DetailKind, librefDef, interval)
  check_args( workgroups, projectId, ccvs, hierarchyMapping, DetailKind, librefDef, interval,
                 {["fn"]="detail_metadata_new", "table", "number", "string", "string", "string",{"string", 'nil'}, {"string", 'nil'}} )
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return 
   end                 
  librefDef = librefDef or ""
  interval = interval or "Month"
  hierarchyMapping = string.gsub(hierarchyMapping, '(\")', '\\%1')
  local pp = [[
  {
    "namespace" : "@DetailKind@",
    "name" : "@wrkgrp@",
    "description" : "Detail Data load information",
    "properties" : [{
        "key" : "parentId",
        "value" : "@parentId@"
      }, {
        "key" : "ccvs",
        "value" : "@ccvs@"
      },
      {
        "key" : "lastStage",
        "value" : "initial"
      }  ,
      {        
        "key" : "stage_initial",
        "value" : "['name':'initial','location':'Future home of librefDef']"
      } ,
      {        
        "key" : "hierarchyMapping",
        "value" : "@hierarchyMapping@"
      },
      {        
        "key" : "interval",
        "value" : "@interval@"
      }
    ]
  } 
  ]]
  local pass, json, alreadyExists, idExists, existingWgrps
  local parentId =""
  
    -- first check for existence of this key combo
    alreadyExists, json = rest.request("get", "/projects/"..projectId.."/properties?name="..workgroups[1].."&namespace="..DetailKind)
    if alreadyExists then 
       local get_id = Json:decode(json )
       existingWgrps = get_id
       if #get_id.items > 0 then
         print(table.tostring(get_id)) 
         parentId = M.metadata_find_entry(get_id.items[1], "parentId")
         if parentId == "" or parentId == "@parentId@" then     
            parentId = get_id.items[1].id 
         end   
         idExists = true   
      else
         idExists = false
      end
  end
    
    
  for i, wrkgrp in ipairs(workgroups) do
    alreadyExists = false
    local wkgrpId
    
    if idExists then -- check to see if we already have recieved this workgroup 
       for j,wg in ipairs(existingWgrps.items) do
          if wg.name == wrkgrp then
             wkgrpId = wg.id 
             alreadyExists = true
             break
          end
       end
    end
    
    local post = string.resolve(pp)
    Debug(post)
    if alreadyExists then
      pass, json = rest.request("put", "/projects/"..projectId .. "/properties/"..wkgrpId, post, nil,"Application/json")
    else
      pass, json = rest.request("post", "/projects/"..projectId .. "/properties", post, nil,"Application/json")
    end
    if not pass then 
       print("Posting or puting meta data failed")
       print(json)
       return false
    end
    if i == 1 then 
       local get_id = Json:decode(json )
       parentId = get_id.id
    end
    -- save and return only the first Id
  end
   
   return true, parentId
  
 end   
 
--- Update detail metadata. This function adds a new stage if necessary and updates "laststage".
--  @param workgroups a table that contains a list of the work groups that the current user belongs to.
--  @param projectId the numeric identifier for the project that contains the detail data.
--  @param DetailKind  the SAS name that indicates the type of detail data. Data with the same variable-to-hierarchy 
--  mapping should be the same kind. 
--  @param stage the name of the data set that corresponds to the processing stage of the detail data (for example,
--  initial, postrules).
--  @return A Boolean value that indicates whether the load ID was created. A value of true indicates success.
--  @return `loadId`: the load ID that was generated for the update.
function M.detail_metadata_update(workgroups, projectId, DetailKind, stage)
  Trace("detail_metadata_update", workgroups, projectId, DetailKind, stage)
  check_args( workgroups, projectId, DetailKind, stage, 
                 {["fn"]="detail_metadata_update", "table", "number", "string", "string"} )
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return 
   end                                
  local pass, json, alreadyExists 
  local parentId
  local get_id
  for i, wrkgrp in ipairs(workgroups) do
    -- first check for existence of this key combo
    local alreadyExists, json = rest.request("get", "/projects/"..projectId.."/properties?name="..wrkgrp.."&namespace="..DetailKind)
    if alreadyExists then 
       get_id = Json:decode(json )
       Debug(get_id)
       if #get_id.items > 0 then 
         parentId = M.metadata_find_entry(get_id.items[1], "parentId")
         if parentId == "" then     
            parentId = get_id.items[1].id
         end
       else
--          sas.print("%1zMissing metadata to update workgroup %s, project %d, detailkind %s.", wrkgrp, projectId, DetailKind)
         sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_detail_metadata_update_1', 'NOQUOTE'), wrkgrp, projectId, DetailKind)
         return false
       end         
    else
--        sas.print("%1zMissing metadata to update workgroup %s, project %d, detailkind %s.", wrkgrp, projectId, DetailKind)
       sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_detail_metadata_update_2', 'NOQUOTE'), wrkgrp, projectId, DetailKind)
       return false
    end
    
    -- check for the existance of this stage
    local stage_name
    local exists =false
    for i, p in ipairs(get_id.items[1].properties) do
       -- look for stage*
       if string.starts_with(p.key, "stage") then
          stage_name = string.match(p.value, "name'%:'(%a+)'")
          if stage_name == stage then
             exists = true
             Debug("Found the stage")
             break
          end
       end
    end
    
    local stage_key = "stage_"..stage
    if not exists then --add it
       local new_stage = { key = stage_key , value = "['name':'"..stage.."','location':'']" }
       get_id.items[1].properties[#get_id.items[1].properties + 1] = new_stage
    end     
   
   -- Update the last active stage
   for i, p in ipairs(get_id.items[1].properties) do
       -- look for last_stage*
       if p.key == "last_stage"  then
          get_id.items[1].properties[i].value = stage_key
       end
   end
    
    -- drop some fields for the post
    get_id.items[1].updatedTimestamp=nil
    get_id.items[1].projectKey=nil
    get_id.items[1].createdByUser=nil
    get_id.items[1].createdTimestamp=nil
    get_id.items[1].updatedByUser=nil
    local current_id = get_id.items[1].id
    get_id.items[1].id=nil
    
    local post = Json:encode(get_id.items[1])
    Debug(post)
    
    -- post the change
    pass, json = rest.request("put", "/projects/"..projectId .. "/properties/"..current_id, post, nil,"Application/json")
   
    if not pass then 
       print("Updating meta data failed")
       print(json)
       return false
    end
    
  end
   
   return true, parentId
  
 end 
 
--- Retrieve detail metadata for a specified work group. 
--  @param workgroup the name of a work group that the current user belongs to.
--  @param projectId the numeric identifier for the project that contains the detail data.
--  @param detailKind the SAS name that indicates the type of detail data. Data with the same variable-to-hierarchy 
--  mapping should be the same kind. 
--  @return A Boolean value that indicates whether the load ID was created. A value of true indicates success.
--  @return `loadId`: the load ID that was used for the data.
--  @return `ccvs`: a string that contains a list of the cross-classification variables, separated by spaces.
--  @return `last_stage`: the name of the data set that corresponds to the most recent processing stage of the detail data.
--  @return `hierarchyMapping`: a JSON string that maps hierarchy level names to cross-classification variables.
--  For example: `{'loanclass': {'loanType':'loantype'}, 'region_state_dim': {'region':'region', 'state':'state'}}`
--  @return `interval`: the time period that is used by the detail data.
function M.detail_metadata_get(workgroup, projectId, detailKind)
  Trace("detail_metadata_get", workgroup, projectId, detailKind)
  check_args( workgroup, projectId, detailKind,
                 {["fn"]="detail_metadata_get", "string", "number", "string"} )
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return 
   end
  local parentId, ccvs, last_stage, hierarchyMapping, interval
  local get_id
  detailKind = detailKind or "default"
  
    --  check for existence of this key combo
  local exists, json = rest.request("get", "/projects/"..projectId.."/properties?name="..workgroup.."&namespace="..detailKind)
  if exists then 
     get_id = Json:decode(json )
     Debug(get_id)
     if not get_id.items[1] then
        -- look any for this kind
        local exists, json = rest.request("get", "/projects/"..projectId.."/properties?namespace="..detailKind)
        if exists then 
          get_id = Json:decode(json )
          Debug(get_id)
        else
          return false
        end
     end
     if not get_id.items[1] then return true end -- No data loaded yet for this wrkgrp
     parentId = M.metadata_find_entry(get_id.items[1], "parentId")
     if parentId == "" then     
        parentId = get_id.items[1].id
     end
     ccvs = M.metadata_find_entry(get_id.items[1], "ccvs")
     last_stage = M.metadata_find_entry(get_id.items[1], "lastStage")
     hierarchyMapping = M.metadata_find_entry(get_id.items[1], "hierarchyMapping")
     interval = M.metadata_find_entry(get_id.items[1], "interval")
  else
--      sas.print("%1zMissing metadata to for workgroup %s, project %d, detailkind %s.", workgroup, projectId, detailKind)
     sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_detail_metadata_get_1', 'NOQUOTE'), workgroup, projectId, detailKind)
     return false
  end
   
   return true, parentId, ccvs, last_stage, hierarchyMapping, interval
  
 end 
 
--- Retrieve detail metadata for a specified load ID.
--  @param loadId the numeric identifier for the specific transfer of detail data into the system.     
--  @param projectId the numeric identifier for the project that contains the detail data.
--  @return A Boolean value that indicates whether the load ID was created. A value of true indicates success.
--  @return `detailKind`: the SAS name that indicates the type of detail data.
--  @return `hierarchyMapping`: a JSON string that maps hierarchy level names to cross-classification variables.
--  For example: `{'loanclass': {'loanType':'loantype'}, 'region_state_dim': {'region':'region', 'state':'state'}}`
--  @return `interval`: the time period that is used by the detail data.
--  @return `location`: the full path to the detail data. This value is returned only when SAS Risk and Finance Workbench
--  is used with SAS Regulatory Risk Management or SAS Risk Management for Banking.
function M.detail_metadata_ldId_get(loadId, projectId)
  Trace("detail_metadata_ldId_get", loadId, projectId)
  check_args( loadId, projectId,
                 {["fn"]="detail_metadata_get", "number", "number"} )
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return 
   end
  local ccvs, last_stage, hierarchyMapping, interval, detailKind, location
  local get_id
  
    --  check for existence of this key combo
  local exists, json = rest.request("get", "/projects/"..projectId.."/properties/"..loadId)
  if exists then 
     get_id = Json:decode(json )
     print(table.tostring(get_id))
     detailKind = get_id.namespace
     ccvs = M.metadata_find_entry(get_id, "ccvs")
     last_stage = M.metadata_find_entry(get_id, "lastStage")
     hierarchyMapping = M.metadata_find_entry(get_id, "hierarchyMapping")
     interval = M.metadata_find_entry(get_id, "interval")
     location = M.metadata_find_entry(get_id, "Location")  -- For RRM and RMB Integration
  else
--      sas.print("%1zMissing metadata to for loadId %d and project %d.", loadId, projectId, detailKind)
     sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_detail_metadata_ldId_get_1', 'NOQUOTE'), loadId, projectId)
     return false
  end
   
   return true, detailKind, hierarchyMapping, interval, location
  
 end                                 

--- Higher Level SAS Risk and Finance Workbench Functions
-- @section RFW



--- Retrieve tagged data from SAS Risk and Finance Workbench.
--  @param projectId the numeric identifier for the project that contains the data to retrieve.
--  @param loadId the numeric identifier for the specific load operation that was used to load the data.     
--  @param ccvs a string that contains a list of the cross-classification variables, separated by spaces.
--  @param hierarchyMapping  a table that maps hierarchy level names to cross-classification variables. For example:
--  `{ region_state_dim = { region = "region", state="state"}, loanclass{ loanType = "loantype"}}`.
--  @param outputDS the output data set to use to store the retrieved data.
--  @return A Boolean value that indicates whether the data was retrieved. A value of true indicates success.
--  @return `workgroupsChanged`: a list of work groups that were changed.
function M.tagged_data_get(  projectId, loadId, ccvs, hierarchyMapping, outputDS )
  Trace("tagged_data_get", projectId, loadId, ccvs, hierarchyMapping, outputDS)
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return 
   end
  local json, pass
  pass, json = rest.request( "get", "projects/".. projectId .. "/taggedData?propertyId=" .. loadId )
  if not pass then
--         sas.print("%1zCan't find property for tagged data mapping for loadid %d.", loadId)
        sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_tagged_data_get_1', 'NOQUOTE'), loadId)
        return false
  end
  
  local dtls = Json:decode(json  )
  Debug(hierarchyMapping)
  Debug(dtls.items)
  
  if #dtls.items == 0 then 
--         sas.print("%1zNo data found for loadid %d.", loadId)
        sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_tagged_data_get_2', 'NOQUOTE'), loadId)
        return false
  end
  
      
    -- create the data set
    local vars = { value = {type="N"}, start_date = {type="C", length=1024}, Entity = {type="C", length=1024}, scenario_name = {type="C", length=64}}
    local c = string.split(ccvs," ")
    for i, cc in ipairs(c) do
       vars[cc] = {type="C", length=1024}
    end
    
    local ds = {}
    local ccv
    ds.vars = vars
    local bucketCode, interval

    for i, dp in ipairs(dtls.items) do
        local row = {}
        ds[#ds+1] = row
        
        if i == 1 then 
           interval = dp.periodType
        end
        row.value = dp.value
        row.start_date = dp.periodStartDate
        row.scenario_name = dp.scenarioName
        -- look up likeys
        for j, h in ipairs(dp.keys) do
          for k, lv in ipairs(h.levels) do
            ccv = hierarchyMapping[h.hierarchy].classificationVarNames[k] 
            if ccv == nil then 
--                sas.print("%1zCan't map ccvs to hierarchy %s.", h.hierarchy)
               sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_tagged_data_get_3', 'NOQUOTE'), h.hierarchy)
               return false
            end
            if lv.member == "" then 
               row[ccv] = "+"    
            else
               row[ccv] = lv.member    
            end
          end
            
        end      
    end
    Debug(ds)

    sas.write_ds(ds, outputDS)
    
     -- fix the date buckets.
     bucketCode = bucket_code(interval)
     sas.submit( [[ data @outputDS@; set @outputDS@;
                  evaldate = input( start_date, anydtdte11.);
                   
                  date_bucket = @bucketCode@;
                  drop evaldate start_date;
                run;]])
  
  return true
end

--- Allocate changes in data from a list of work groups in SAS Risk and Finance Workbench. Allocation distributes changes to a subtotal 
--  value in a cell across all of the child line items, according to a specified allocation method.
--  @param projectId the numeric identifier for the project that contains the data to allocate.
--  @param stage the name of the data set that corresponds to the processing stage of the detail data to allocate. 
--  @param detailKind the SAS name that indicates the type of detail data. Data with the same variable-to-hierarchy mapping should be the same kind.
--  @param allocationRule[opt] the allocation method to use for distributing subtotal values. The following methods are supported:
--
--   * `proportional`: The new subtotal value is distributed across the child line items in proportion to the contribution of 
--     each line item to the subtotal.
--   * `even`: The increase or decrease in the subtotal value is evenly distributed across the child line items. Each line-item value 
--     is modified by the same amount.
-- 
--  The default value is `proportional`.
--  @param mgversion[opt] the version of the model group to use for the allocation operation. The default value is `""`. 
--  @return A Boolean value that indicates whether the data was allocated. A value of true indicates success.
--  @return `workGrps`: a list of the work groups for which data was changed.
function M.allocate_data(  projectId, stage, detailKind, allocationRule, mgversion )
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return 
   end
   Trace("allocate_data", projectId, stage, detailKind, allocationRule, mgversion)

   local begintime = sas.time()
   
   mgversion = mgversion or ""
   
   local scenario_names = M.scenario_names_get(projectId)
   if scenario_names ~= nil and #scenario_names == 0 then
      scenario_names = nil
   end
   Debug("From scenario_names_get")
   Debug(scenario_names)

     -- Get a total summary data set from all the workgroups for the most recent data
   local pass, wg, loadId, ccvs, last_stage, hierarchyMapping = M.detail_data_sum(projectId, "work.summary", detailKind)
   Debug("Last Stage=",last_stage)
   local pass = M.tagged_data_get(  projectId, loadId, ccvs, hierarchyMapping, "work.sheet" )

   if not pass then 
--        sas.print("%1zCouldn't get tagged data from project.")
       sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_allocate_data_1', 'NOQUOTE'))
       return false
   end
   
   -- Now do the comparison. First remove summary rows that weren't returned from the 
   -- query or that match the query
   -- We are left with a data set that contains only differences.
   sas.submit([[
     /* put a key on the old file and new file */
     data work.summary(compress=yes); set work.summary;
        length _id $ 10000;
        array _temp[*] @ccvs@ date_bucket entity scenario_name;
        
        _id="";
        do i=1 to dim(_temp);
           _id = trim(_id) || trim(_temp[i]);
        end;
        _id = md5(_id); 
         
        drop i;
     run;

     data work.drop(compress=yes); set work.sheet;
        length _id $ 10000;
        array _temp[*] @ccvs@ date_bucket entity scenario_name;
        
        _id="";
        do i=1 to dim(_temp);
           _id = trim(_id) || trim(_temp[i]);
        end;
        _id = md5(_id);
         
        new_value = value;
        keep _id new_value;
        
     run;
     data work.summary; set work.summary;
        length _id $ 10000 new_value 8;
        if _n_ = 1 then do;
          declare hash myhash (dataset: "work.drop"); 
          rc = myhash.defineKey("_id"); 
          rc = myhash.defineData("new_value");
          rc = myhash.defineDone(); 
        end;

        rc = myhash.find();
        reldif = abs(new_value - value)/(abs(value) + 1.0e-5);
        dif = new_value - value;
      
        if rc =  0 and abs(reldif) > 1.0e-8 then output;
        drop _id rc;
      run;
  
     ]])
     
     
      --  Begin the allocation process
      local differences = sas.read_ds("work.summary")
      if #differences == 0 then
          -- nothing to disaggregate/allocate
--           sas.print("%3zNo differences to allocate.")
          sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_allocate_data_2', 'NOQUOTE'))
          return true

      end      
      
      -- loop over differences
      -- create code that passes over the data just once and does the adjustments
      local ccvs_plus = ccvs .. " date_bucket entity scenario_name"
    
      local code = "data work.details(compress=yes); set work.details;\n"
      allocationRule = allocationRule or "proportional"
      for i, dif in ipairs(differences) do
         
      -- check the difference
         if  dif.dif ~= nil  then
             if allocationRule == "even" then
                code = code .. "\n" .. alloc.apply_equal( ccvs_plus, dif, dif.dif)
             else -- rule is proportional
                code = code .. "\n" ..alloc.apply_relative( ccvs_plus, dif, dif.dif)
             end
         end
      end
      code = code .. "\nrun;"
   
   
  -- loop over the list of workgroups and perform allocation to the detail data
    local success, workGrps =  M.detail_data_get( projectId, detailKind, "work.details" )
   
   -- Apply the allocations
    sas.submit(code)
           
    -- Update the final detail data set
    M.detail_data_put( projectId,  "work.details", stage, nil, detailKind, nil, nil, nil, mgversion)

  -- Now ready to re-sum the numbers.
  local pass = M.agg_data( projectId, stage, detailKind )
  
--   sas.print("%3zTotal allocation time %d seconds.", sas.time()-begintime)
  
  sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_allocate_data_3', 'NOQUOTE'), sas.time()-begintime)
  return pass, workGrps

end

---Aggregate data from a list of work groups in SAS Risk and Finance Workbench.
--  @param projectId the numeric identifier for the project that contains the data to aggregate.
--  @param stage[opt] the name of the data set that corresponds to the processing stage of the detail data to aggregate. 
--  The default value is the data set for the last stage.
--  @param detailKind[opt] the SAS name that indicates the type of detail data to aggregate. Data with the same variable-to-hierarchy 
--  mapping should be the same kind.
--  @return A Boolean value that indicates whether the data was aggregated. A value of true indicates success.
function M.agg_data( projectId, stage, detailKind )
   if not M.url_set() then  
      sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_url_not_set', 'NOQUOTE'))
      return 
   end
  Trace("agg_data", projectId, stage, detailKind )

  -- Summarize the data
  local pass, wgrps, property_id, ccvs, latest_stage = M.detail_data_sum(projectId, "work.summary", detailKind)
   
  if not pass then
--      sas.print("%1zSumarization failed.")
     sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_agg_data_1', 'NOQUOTE'))
     return false
  end
  stage = stage or latest_stage
  property_id = tonumber(property_id)  
  
  --  copy to out_lib.results for the correct scenario_name and entity
  --  First get list of entities and scenario names
  local scenarios, entities
  if alloc.var_exists("work.summary","scenario_name","entity") then 
    sas.submit([[   proc sql noprint;
                      select distinct entity, scenario_name into :ent_name separated by ','  , :scn_name separated by ',' from work.summary;
                    quit;
           ]])
    scenarios = string.split(sas.symget("scn_name"), ',')
    entities = string.split(sas.symget("ent_name"), ',')
   else
   
    sas.submit([[   proc sql noprint;
                    select distinct entity into :ent_name separated by ',' from work.summary;
                  quit;
         ]])
     entities = string.split(sas.symget("ent_name"), ',')
     scenarios = {}
     for i,e in ipairs(entities) do
         scenarios[i] = ""
     end
   end
    
   -- loop over scenarios and entities 
   if M.testing then return true end
   for i, e in ipairs(entities) do
     local s = scenarios[i]
     local sbar=nil 
     if s ~= "" then sbar = s end
     local mg_prop = M.model_group_run_get_properties(projectId,"Load_Details_"..detailKind,e,sbar)
      print(table.tostring(mg_prop))
     if not mg_prop then
--         sas.print("%1zModel group run missing for entity %s and scenario %s.", e, s)
        sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_agg_data_2', 'NOQUOTE'), e, s)
        return false
     end
     local out_lib = mg_prop.outputDir
     e=string.gsub(e,"'","''")
     local where_cls = "where entity='"..e.."'"
     if s ~= "" then
        where_cls = where_cls .. "and scenario_name='"..s.."'"
     end
     sas.submit([[  
           libname out_lib "@out_lib@";            
           data out_lib.results; set work.summary;
               @where_cls@;
               evaldate=input(date_bucket, date.);
               property_id = @property_id@;
           run;               
           ]] )
               
     -- Load data into RFW and reconcile the loaded data with the summary data    
      local updatedTimestamp = sas.put(sas.datetime(),"D15.")      

      if not M.model_group_data_load(mg_prop.runKey,updatedTimestamp) then
--        sas.print("%1zUnable to complete aggregation process." )
       sas.print(sas.sasmsg('sashelp.stressluamsg',  'sst_agg_data_3', 'NOQUOTE' ))
       return false
      end

     end
     
     return true
end


return M
 
